import copy
import math
from itertools import product
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import cast

import attr
import networkx as nx
import numpy as np
import scipy.ndimage
from matplotlib import pyplot as plt
from networkx import Graph
from scipy.spatial.transform import Rotation

from avalon.common.errors import SwitchError
from avalon.contrib.serialization import Serializable
from avalon.datagen.godot_base_types import IntRange
from avalon.datagen.world_creation.geometry import BuildingTile
from avalon.datagen.world_creation.indoor.blocks import FloorBlock
from avalon.datagen.world_creation.indoor.blocks import LadderBlock
from avalon.datagen.world_creation.indoor.blocks import LevelBlock
from avalon.datagen.world_creation.indoor.constants import CEILING_THICKNESS
from avalon.datagen.world_creation.indoor.constants import DEFAULT_FLOOR_THICKNESS
from avalon.datagen.world_creation.indoor.constants import LADDER_THICKNESS
from avalon.datagen.world_creation.indoor.constants import MIN_BUILDING_SIZE
from avalon.datagen.world_creation.indoor.constants import SLOPE_THICKNESS
from avalon.datagen.world_creation.indoor.constants import TILE_SIZE
from avalon.datagen.world_creation.indoor.constants import Azimuth
from avalon.datagen.world_creation.indoor.constants import TileIdentity
from avalon.datagen.world_creation.indoor.tiles import draw_line_in_grid
from avalon.datagen.world_creation.types import BuildingBoolNP
from avalon.datagen.world_creation.types import BuildingFloatNP
from avalon.datagen.world_creation.types import BuildingIntNP
from avalon.datagen.world_creation.types import Point3DNP


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Room:
    id: int
    position: BuildingTile
    width: int
    length: int
    outer_height: float
    floor_heightmap: BuildingFloatNP = attr.field(
        eq=attr.cmp_using(eq=np.array_equal),  # type: ignore[attr-defined]
        default=attr.Factory(lambda self: np.ones((self.length, self.width), dtype=np.float_), takes_self=True),
    )
    climbable_mask: BuildingBoolNP = attr.field(
        eq=attr.cmp_using(eq=np.array_equal),  # type: ignore[attr-defined]
        default=attr.Factory(lambda self: np.zeros((self.length, self.width), dtype=np.bool_), takes_self=True),
    )

    def with_heightmap(self, new_heightmap: BuildingFloatNP) -> "Room":
        return attr.evolve(self, floor_heightmap=new_heightmap)

    @property
    def tiles(self) -> BuildingIntNP:
        return np.full((self.length, self.width), TileIdentity.ROOM.value, dtype=np.int_)

    def get_tile_positions(self, buffer: int = 0) -> Tuple[Tuple[int, int], ...]:
        """
        Returns (z, x) tuples of building-space coordinates that constitute this room, optionally including a buffer
        around the room to include its wall-space. Naive - doesn't account for values that would be OOB in the story.
        """
        return tuple(
            product(
                range(self.position.z - buffer, self.position.z + self.length + buffer),
                range(self.position.x - buffer, self.position.x + self.width + buffer),
            )
        )

    def with_position(self, new_position: BuildingTile) -> "Room":
        return attr.evolve(self, position=new_position)

    def with_id(self, new_id: int) -> "Room":
        return attr.evolve(self, id=new_id)

    @property
    def floor_space(self) -> int:
        return self.length * self.width

    @property
    def floor_negative_depth(self) -> float:
        return cast(float, abs(min(0, self.floor_heightmap.min())))

    @property
    def center(self) -> BuildingTile:
        return BuildingTile(x=self.position.x + (self.width // 2), z=self.position.z + (self.length // 2))

    @property
    def z_range(self) -> IntRange:
        return IntRange(self.position.z, self.position.z + self.length)

    @property
    def x_range(self) -> IntRange:
        return IntRange(self.position.x, self.position.x + self.width)


class RoomPlacementError(Exception):
    pass


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Hallway(Serializable):
    """
    Hallway points  include joint points with rooms, so we can easily tell if the joint is horizontal / vertical.
    Points MUST be in the correct order (from_room, to_room) when creating the Hallway.
    """

    points: Tuple[BuildingTile, ...]
    from_room_id: int
    to_room_id: int
    width: int
    height: Optional[int] = None  # None => entire room height

    @property
    def total_length(self) -> int:
        length = 0
        for point_a, point_b in zip(self.points[:-1], self.points[1:]):
            if point_a.x == point_b.x:
                segment_length = abs(point_a.z - point_b.z)
            else:
                segment_length = abs(point_a.x - point_b.x)
            length += segment_length
        return length

    @property
    def from_room_azimuth(self) -> Azimuth:
        """What direction is the entryway into the hallway pointing to, when going from the 'from' room?"""
        return self._azimuth(self.points[0], self.points[1])

    @property
    def to_room_azimuth(self) -> Azimuth:
        """What direction is the entryway into the hallway pointing to, when going from the 'to' room?"""
        return self._azimuth(self.points[-1], self.points[-2])

    def _azimuth(self, point_a: BuildingTile, point_b: BuildingTile) -> Azimuth:
        if point_a.x == point_b.x:
            return Azimuth.SOUTH if point_a.z < point_b.z else Azimuth.NORTH
        elif point_a.z == point_b.z:
            return Azimuth.EAST if point_a.x < point_b.x else Azimuth.WEST
        else:
            raise SwitchError((point_a, point_b))


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class StoryLink(Serializable):
    bottom_story_id: int
    bottom_room_id: int
    bottom_position: BuildingTile  # in story space
    top_story_id: int
    top_room_id: int
    top_position: BuildingTile  # in story space

    def get_link_length(self, bottom_story: "Story", top_story: "Story") -> float:
        raise NotImplementedError

    def get_level_blocks(self, bottom_story: "Story", top_story: "Story") -> Sequence[LevelBlock]:
        raise NotImplementedError


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Ramp(StoryLink):
    width: int  # width will be expanded right+down from specified positions

    def get_link_length(self, bottom_story: "Story", top_story: "Story") -> float:
        run = self.slope_run
        rise = bottom_story.outer_height + top_story.floor_negative_depth
        return math.sqrt(run**2 + rise**2)

    @property
    def slope_run(self) -> float:
        # start/end positions are the landings - the actual slope floor length does not include these
        if self.bottom_position.x == self.top_position.x:
            return abs(self.bottom_position.z - self.top_position.z) - 1
        else:
            return abs(self.bottom_position.x - self.top_position.x) - 1

    def _get_geometry(self, bottom_story: "Story", top_story: "Story") -> Tuple[float, ...]:
        # TODO: 1d9f913a-3870-41b0-839e-8b29d0b89983
        bottom_room = bottom_story.rooms[self.bottom_room_id]
        bottom_position_in_room_space = BuildingTile(
            self.bottom_position.x - bottom_room.position.x, self.bottom_position.z - bottom_room.position.z
        )
        bottom_story_floor_positive_depth = (
            bottom_room.floor_heightmap[bottom_position_in_room_space.z, bottom_position_in_room_space.x]
            - DEFAULT_FLOOR_THICKNESS
        )
        bottom_story_floor_depth = bottom_story.floor_negative_depth + bottom_story_floor_positive_depth

        top_room = top_story.rooms[self.top_room_id]
        top_position_in_room_space = BuildingTile(
            self.top_position.x - top_room.position.x, self.top_position.z - top_room.position.z
        )
        top_story_floor_depth = (
            top_story.floor_negative_depth
            + top_room.floor_heightmap[top_position_in_room_space.z, top_position_in_room_space.x]
            - DEFAULT_FLOOR_THICKNESS
        )

        slope_width = float(self.width)
        slope_run = float(self.slope_run)
        slope_rise = (bottom_story.outer_height - bottom_story_floor_positive_depth) + top_story_floor_depth
        slope_length = math.sqrt(slope_run**2 + slope_rise**2)
        slope_angle = math.atan(slope_rise / slope_run)
        slope_thickness = SLOPE_THICKNESS

        complementary_angle = math.radians(90) - slope_angle
        # Our geometry is based on placing a plane, but we're placing a block, so we need to account for thickness
        thickness_y_offset = (slope_thickness / 2) * math.sin(complementary_angle)
        thickness_xz_offset = (slope_thickness / 2) * math.cos(complementary_angle)

        is_horizontal = self.bottom_position.z == self.top_position.z
        # +1s are to offset the slope from its bottom landing
        if is_horizontal:
            plane_width = slope_length
            plane_length = slope_width
            angle_multiplier = 1 if self.top_position.x > self.bottom_position.x else -1
            pitch, roll = 0.0, angle_multiplier * math.degrees(slope_angle)
            x = min(self.bottom_position.x, self.top_position.x) + (slope_run / 2) + 1 + thickness_xz_offset
            z = min(self.bottom_position.z, self.top_position.z) + (slope_width / 2)
        else:
            plane_width = slope_width
            plane_length = slope_length
            angle_multiplier = -1 if self.top_position.z > self.bottom_position.z else 1
            pitch, roll = angle_multiplier * math.degrees(slope_angle), 0.0
            x = min(self.bottom_position.x, self.top_position.x) + (slope_width / 2)
            z = min(self.bottom_position.z, self.top_position.z) + (slope_run / 2) + 1 + thickness_xz_offset

        y = math.tan(slope_angle) * (slope_run / 2) - thickness_y_offset + bottom_story_floor_depth
        return x, y, z, plane_width, slope_thickness, plane_length, pitch, 0, roll

    def get_level_blocks(self, bottom_story: "Story", top_story: "Story") -> List[LevelBlock]:
        x, y, z, plane_width, slope_thickness, plane_length, pitch, yaw, roll = self._get_geometry(
            bottom_story, top_story
        )
        return [
            FloorBlock.make(
                np.array([x, y, z]),
                np.array([plane_width, slope_thickness, plane_length]),
                rotation=Rotation.from_euler("xyz", [pitch, yaw, roll], degrees=True),
            )
        ]


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Ladder(StoryLink):
    width: int
    azimuth: Azimuth

    def get_link_length(self, bottom_story: "Story", top_story: "Story") -> float:
        return bottom_story.outer_height + top_story.floor_negative_depth

    def get_geometry(self, bottom_story: "Story", top_story: "Story") -> Tuple[float, ...]:
        # todo: dedup
        bottom_room = bottom_story.rooms[self.bottom_room_id]

        bottom_position_in_room_space = BuildingTile(
            self.bottom_position.x - bottom_room.position.x, self.bottom_position.z - bottom_room.position.z
        )
        bottom_story_floor_positive_depth = (
            bottom_room.floor_heightmap[bottom_position_in_room_space.z, bottom_position_in_room_space.x]
            - DEFAULT_FLOOR_THICKNESS
        )
        bottom_story_floor_depth = bottom_story.floor_negative_depth + bottom_story_floor_positive_depth

        top_room = top_story.rooms[self.top_room_id]
        top_position_in_room_space = BuildingTile(
            self.top_position.x - top_room.position.x, self.top_position.z - top_room.position.z
        )
        top_story_floor_depth = (
            top_story.floor_negative_depth
            + top_room.floor_heightmap[top_position_in_room_space.z, top_position_in_room_space.x]
            - DEFAULT_FLOOR_THICKNESS
        )

        tile_size = float(TILE_SIZE)
        ladder_height = (bottom_story.outer_height - bottom_story_floor_positive_depth) + top_story_floor_depth
        ladder_thickness = LADDER_THICKNESS

        y = bottom_story_floor_depth + ladder_height / 2
        if self.azimuth == Azimuth.NORTH:
            x, z = self.top_position.x + tile_size / 2, self.top_position.z + ladder_thickness / 2
            ladder_width, ladder_length = tile_size, ladder_thickness
        elif self.azimuth == Azimuth.EAST:
            x, z = self.top_position.x + tile_size - ladder_thickness / 2, self.top_position.z + tile_size / 2
            ladder_width, ladder_length = ladder_thickness, tile_size
        elif self.azimuth == Azimuth.SOUTH:
            x, z = self.top_position.x + tile_size / 2, self.top_position.z + tile_size - ladder_thickness / 2
            ladder_width, ladder_length = tile_size, ladder_thickness
        elif self.azimuth == Azimuth.WEST:
            x, z = self.top_position.x + ladder_thickness / 2, self.top_position.z + tile_size / 2
            ladder_width, ladder_length = ladder_thickness, tile_size
        else:
            raise SwitchError(f"Unexpected Azimuth {self.azimuth}")
        return x, y, z, ladder_width, ladder_height, ladder_length

    def get_level_blocks(self, bottom_story: "Story", top_story: "Story") -> Sequence[LevelBlock]:
        x, y, z, ladder_width, ladder_height, ladder_length = self.get_geometry(bottom_story, top_story)

        collision_block = LadderBlock.make(
            np.array([x, y, z]),
            np.array([ladder_width, ladder_height, ladder_length]),
            rotation=Rotation.identity(),
            is_visual=False,
            is_collider=True,
            is_climbable=True,
        )

        tile_size = 1
        rail_thickness = LADDER_THICKNESS
        if self.azimuth in {Azimuth.NORTH, Azimuth.SOUTH}:
            left_rail_x, left_rail_z = x - tile_size / 2 + rail_thickness / 2, z
        else:
            left_rail_x, left_rail_z = x, z - tile_size / 2 + rail_thickness / 2
        left_rail = LadderBlock.make(
            np.array([left_rail_x, y, left_rail_z]),
            np.array([rail_thickness, ladder_height, rail_thickness]),
            rotation=Rotation.identity(),
            is_visual=True,
            is_collider=False,
        )

        if self.azimuth in {Azimuth.NORTH, Azimuth.SOUTH}:
            right_rail_x, right_rail_z = x + tile_size / 2 - rail_thickness / 2, z
        else:
            right_rail_x, right_rail_z = x, z + tile_size / 2 - rail_thickness / 2
        right_rail = LadderBlock.make(
            np.array([right_rail_x, y, right_rail_z]),
            np.array([rail_thickness, ladder_height, rail_thickness]),
            rotation=Rotation.identity(),
            is_visual=True,
            is_collider=False,
        )

        step_size = 0.4
        step_thickness = LADDER_THICKNESS
        step_count, margin = divmod(ladder_height, step_size)
        if ladder_width > ladder_length:
            step_width = ladder_width - 2 * step_thickness
            step_length = ladder_length
        else:
            step_width = ladder_width
            step_length = ladder_length - 2 * step_thickness
        steps = []
        for i in range(int(step_count)):
            steps.append(
                LadderBlock.make(
                    np.array([x, margin + i * step_size, z]),
                    np.array([step_width, step_thickness, step_length]),
                    rotation=Rotation.identity(),
                    is_visual=True,
                    is_collider=False,
                )
            )

        visual_blocks = [left_rail, right_rail, *steps]

        return [collision_block] + visual_blocks


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Window(Serializable):
    position: Point3DNP  # centroid
    size: Point3DNP
    is_passable: bool = False  # should the cutout have a collision shape


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Entrance(Serializable):
    story_num: int
    connected_room_id: int
    azimuth: Azimuth
    points: Tuple[BuildingTile, ...]
    width: int = 1
    height: int = 3

    exterior_id_by_azimuth = {
        Azimuth.NORTH: -1,
        Azimuth.SOUTH: -2,
        Azimuth.WEST: -3,
        Azimuth.EAST: -4,
    }

    @property
    def hallway(self) -> Hallway:
        return Hallway(
            self.points, Entrance.exterior_id_by_azimuth[self.azimuth], self.connected_room_id, self.width, self.height
        )

    def get_points_in_story_outline(self, story: "Story") -> List[Tuple[int, int]]:
        outline_mask = story.get_outline_mask()
        entrance_mask = np.zeros_like(outline_mask)
        draw_line_in_grid(self.points, entrance_mask, TileIdentity.HALLWAY.value, include_ends=False)
        entrance_mask_points_in_outline = np.where(outline_mask & entrance_mask)
        return list(zip(*entrance_mask_points_in_outline))  # type: ignore

    def get_connected_room_and_landing_position(self, story: "Story") -> Tuple[Room, BuildingTile]:
        """note: landing position is returned in room space"""
        room = story.rooms[self.connected_room_id]
        entrance_grid = np.zeros((story.length, story.width))
        draw_line_in_grid(self.points, entrance_grid, TileIdentity.HALLWAY.value, include_ends=True)
        room_grid = np.zeros_like(entrance_grid)
        set_room_tiles_in_grid(room, room.tiles, room_grid)
        entrance_hallway_overlaps_with_room = entrance_grid.astype(np.bool_) & room_grid.astype(np.bool_)
        room_landing_coords = list(zip(*np.where(entrance_hallway_overlaps_with_room)))
        room_landing = list(room_landing_coords[0])
        room_landing[0] -= room.position.z
        room_landing[1] -= room.position.x
        return room, BuildingTile(x=room_landing[1], z=room_landing[0])


@attr.s(auto_attribs=True, eq=False, collect_by_mro=True)
class Story:
    r"""
     ___________
    |__________| \  } ceiling thickness (1), inset into room height
    |          |  \                     \
    |          |   } outer height (5);   } inner height = 5-1-1=3
    |____  ____|  |                     /
    |___| |____| /   } positive floor depth (1), offset INTO room height; bottom line is 0-level
    |   | |    | \
    |   | |    |  } negative floor depth (3)
    |___|_|____| /
    """

    num: int
    width: int = attr.field(validator=attr.validators.ge(MIN_BUILDING_SIZE))
    length: int = attr.field(validator=attr.validators.ge(MIN_BUILDING_SIZE))
    footprint: BuildingBoolNP = attr.field(
        eq=attr.cmp_using(eq=np.array_equal),  # type: ignore[attr-defined]
        default=attr.Factory(lambda self: np.ones((self.length, self.width), dtype=np.bool_), takes_self=True),
    )
    rooms: List[Room] = attr.Factory(list)
    hallways: List[Hallway] = attr.Factory(list)
    story_links: List[StoryLink] = attr.Factory(list)
    windows: List[Window] = attr.Factory(list)
    entrances: List[Entrance] = attr.Factory(list)
    has_ceiling: bool = False

    @property
    def outer_height(self) -> float:
        return max([r.outer_height for r in self.rooms])

    @property
    def inner_height(self) -> float:
        # The EXACT inner height can vary tile by tile depending on terrain; this is the "base" inner height assuming
        # the floor uses DEFAULT_FLOOR_THICKNESS.
        return self.outer_height - DEFAULT_FLOOR_THICKNESS - (CEILING_THICKNESS if self.has_ceiling else 0)

    @property
    def floor_negative_depth(self) -> float:
        # Note that all positive story terrain is projected INTO the room (rather than offsetting the room & walls),
        # see diagram above
        return max([room.floor_negative_depth for room in self.rooms])

    @property
    def floor_heightmap(self) -> BuildingFloatNP:
        """returns positive floor thickness for all walkable tiles, np.nan elsewhere"""
        tiles = np.full((self.length, self.width), np.nan, dtype=np.float32)
        for room in self.rooms:
            set_room_tiles_in_grid(room, room.floor_heightmap, tiles)
        for hallway in self.hallways:
            draw_line_in_grid(
                hallway.points, tiles, DEFAULT_FLOOR_THICKNESS, include_ends=False, drawable_grid_value=None
            )
        return tiles

    def get_room_at_point(self, point: BuildingTile) -> Optional[Room]:
        for room in self.rooms:
            if (
                room.position.x <= point.x <= room.position.x + room.width - 1
                and room.position.z <= point.z <= room.position.z + room.length - 1
            ):
                return room
        return None

    def get_outline_mask(self) -> BuildingBoolNP:
        is_building = self.footprint.astype(np.int8) != 0
        convolution = scipy.ndimage.convolve(self.footprint.astype(np.int8), np.ones((3, 3)), mode="constant")
        return cast(BuildingBoolNP, is_building & (convolution != 9))

    def get_exterior_wall_azimuths_by_room_id(self) -> Dict[int, List[Azimuth]]:
        outline_mask = self.get_outline_mask()
        exterior_wall_azimuths_by_room_id = {}
        for room in self.rooms:
            room_has_exterior_wall_by_azimuth = {
                Azimuth.NORTH: outline_mask[room.position.z - 1, room.position.x],
                Azimuth.SOUTH: outline_mask[room.position.z + room.length, room.position.x],
                Azimuth.EAST: outline_mask[room.position.z, room.position.x + room.width],
                Azimuth.WEST: outline_mask[room.position.z, room.position.x - 1],
            }
            exterior_wall_azimuths_by_room_id[room.id] = [
                azimuth
                for azimuth, has_exterior_wall in room_has_exterior_wall_by_azimuth.items()
                if has_exterior_wall
            ]
        return exterior_wall_azimuths_by_room_id

    def generate_tiles(
        self,
        include_links_going_up: bool = True,
        include_links_going_down: bool = True,
        include_hallway_landings: bool = False,
    ) -> BuildingIntNP:
        tiles = np.invert(self.footprint).astype(np.float32) * TileIdentity.VOID.value
        for room in self.rooms:
            set_room_tiles_in_grid(room, room.tiles, tiles)
        for hallway in self.hallways:
            draw_line_in_grid(hallway.points, tiles, TileIdentity.HALLWAY.value, include_ends=include_hallway_landings)
        for link in self.story_links:
            if link.bottom_story_id == self.num and include_links_going_up:
                set_link_in_grid(link, tiles, set_bottom_landing=True, set_top_landing=False)
            elif link.top_story_id == self.num and include_links_going_down:
                set_link_in_grid(link, tiles, set_bottom_landing=False, set_top_landing=True)
        return tiles

    def generate_room_id_tiles(self) -> BuildingIntNP:
        tiles = np.invert(self.footprint).astype(np.int8) * TileIdentity.VOID.value
        for room in self.rooms:
            set_room_tiles_in_grid(room, np.full((room.length, room.width), room.id), tiles)
        return tiles


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Obstacle:
    story_id: int
    room_id: int

    def apply(self, stories: List[Story]) -> None:
        raise NotImplementedError


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Wall(Obstacle):
    points: Tuple[BuildingTile, ...]
    thickness: int
    height: float

    def apply(self, stories: List[Story]) -> None:
        height = self.height
        room = stories[self.story_id].rooms[self.room_id]
        new_heightmap = room.floor_heightmap.copy()
        draw_line_in_grid(self.points, new_heightmap, height, drawable_grid_value=None)
        deformed_room = room.with_heightmap(new_heightmap)
        stories[self.story_id].rooms[self.room_id] = deformed_room


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class FloorChasm(Obstacle):
    site_mask: BuildingBoolNP
    climbable_mask: BuildingBoolNP
    depth_map: BuildingFloatNP

    @property
    def chasm_mask(self) -> BuildingBoolNP:
        return self.site_mask & ~np.isnan(self.depth_map)

    def apply(self, stories: List[Story]) -> None:
        room = stories[self.story_id].rooms[self.room_id]
        new_heightmap = room.floor_heightmap.copy()
        new_heightmap[self.chasm_mask] = -self.depth_map[self.chasm_mask]
        new_climbable_mask = room.climbable_mask.copy()
        new_climbable_mask[self.site_mask] = self.climbable_mask[self.site_mask]
        stories[self.story_id].rooms[self.room_id] = attr.evolve(
            room, floor_heightmap=new_heightmap, climbable_mask=new_climbable_mask
        )


class TileGraph(Graph):
    def plot(self) -> None:
        fig, ax = plt.subplots(1, figsize=(10, 10))
        plt.gca().set_aspect("equal")
        plt.gca().invert_yaxis()
        nx.draw_networkx(self, pos={node: tuple(reversed(node)) for node in self.nodes}, ax=ax)


class RoomTileGraph(TileGraph):
    def __init__(self, rooms: Iterable[Room]) -> None:
        super().__init__()
        length = max([r.position.z + r.length for r in rooms])
        width = max([r.position.x + r.width for r in rooms])
        nx.grid_2d_graph(length, width, create_using=self)

    def find_shortest_path(
        self, point_a: BuildingTile, point_b: BuildingTile, avoid_rooms: Tuple[Room, ...] = tuple()
    ) -> List[Tuple[int, int]]:
        graph_copy = copy.deepcopy(self)
        avoid_points: List[Tuple[int, int]] = sum(
            [list(room.get_tile_positions(buffer=1)) for room in avoid_rooms], []
        )
        graph_copy.remove_nodes_from(avoid_points)
        return nx.astar_path(graph_copy, (point_a.z, point_a.x), (point_b.z, point_b.x))  # type: ignore


class StoryNavGraph(TileGraph):
    def __init__(self, story: Story, exclude_exterior: bool = True) -> None:
        super().__init__()
        nx.grid_2d_graph(story.length, story.width, create_using=self)
        tiles = story.generate_tiles()
        # Note: we do not include void below, since you could have unconnected rooms that each have their own entrance
        unwalkable_tile_identities = {TileIdentity.LINK, TileIdentity.FULL}
        if exclude_exterior:
            unwalkable_tile_identities.add(TileIdentity.VOID)
        unwalkable_tiles = list(
            zip(*np.where(np.isin(tiles, [identity.value for identity in unwalkable_tile_identities])))
        )
        self.remove_nodes_from(unwalkable_tiles)


def generate_room(room_id: int, width: int, height: int, length: int) -> Room:
    assert room_id != 0
    return Room(id=room_id, position=BuildingTile(0, 0), width=width, length=length, outer_height=height)


def add_room_tiles_to_grid(room: Room, tiles: BuildingIntNP, grid: BuildingIntNP) -> None:
    local_grid = grid[room.position.z : room.position.z + room.length, room.position.x : room.position.x + room.width]
    local_grid[local_grid == 0] += tiles[local_grid == 0]


def set_room_tiles_in_grid(room: Room, tiles: BuildingIntNP, grid: BuildingIntNP) -> None:
    grid[room.position.z : room.position.z + room.length, room.position.x : room.position.x + room.width] = tiles


def set_link_in_grid(link: StoryLink, grid: BuildingIntNP, set_bottom_landing: bool, set_top_landing: bool) -> None:
    if set_bottom_landing:
        grid[link.bottom_position.z, link.bottom_position.x] = TileIdentity.LINK_BOTTOM_LANDING.value
    if set_top_landing:
        grid[link.top_position.z, link.top_position.x] = TileIdentity.LINK_TOP_LANDING.value
    start = min(link.bottom_position.z, link.top_position.z), min(link.bottom_position.x, link.top_position.x)
    end = max(link.bottom_position.z, link.top_position.z), max(link.bottom_position.x, link.top_position.x)
    if start[0] == end[0]:
        start = start[0], start[1] + 1
        end = end[0], end[1] - 1
    else:
        start = start[0] + 1, start[1]
        end = end[0] - 1, end[1]
    grid[start[0] : end[0] + 1, start[1] : end[1] + 1] = TileIdentity.LINK.value



import multiprocessing.synchronize
import signal
import time
import traceback
from multiprocessing.context import BaseContext
from typing import Any
from typing import Optional
from typing import Tuple

import attr
import gym
import torch
from loguru import logger
from torch import Tensor
from tree import map_structure

from avalon.agent.common.envs import build_env
from avalon.agent.common.get_algorithm_cls import get_algorithm_cls
from avalon.agent.common.params import EnvironmentParams
from avalon.agent.common.params import Params
from avalon.agent.common.storage import EpisodeStorage
from avalon.agent.common.storage import ModelStorage
from avalon.agent.common.storage import TrajectoryStorage
from avalon.agent.common.types import ActionBatch
from avalon.agent.common.types import Algorithm
from avalon.agent.common.types import StepData
from avalon.agent.common.util import create_action_storage
from avalon.agent.common.util import create_observation_storage
from avalon.agent.common.util import get_checkpoint_file
from avalon.agent.common.util import masked_copy_structure
from avalon.agent.common.util import postprocess_uint8_to_float
from avalon.agent.common.util import seed_and_run_deterministically_if_enabled
from avalon.agent.common.util import setup_new_process
from avalon.agent.dreamer.params import OffPolicyParams
from avalon.common.error_utils import capture_exception
from avalon.common.error_utils import setup_sentry
from avalon.common.log_utils import configure_parent_logging


class AsyncRolloutManager:
    def __init__(  # type: ignore
        self,
        params: OffPolicyParams,
        obs_space: gym.spaces.Dict,
        action_space: gym.spaces.Dict,
        rollout_manager_id: int,
        env_step_counter,  # multiprocessing.Value; typing for this is broken
        multiprocessing_context: BaseContext,
        train_rollout_dir: str,
    ) -> None:
        seed_and_run_deterministically_if_enabled()

        self.params = params
        self.obs_space = obs_space
        self.action_space = action_space
        self.rollout_manager_id = rollout_manager_id
        self.multiprocessing_context = multiprocessing_context
        self.env_step_counter = env_step_counter
        self.shutdown_event = multiprocessing_context.Event()
        self.train_rollout_dir = train_rollout_dir
        self.wandb_queue = multiprocessing_context.Queue()

    def start(self) -> None:
        ps = self.multiprocessing_context.Process(
            target=self.safe_entrypoint, args=(self.shutdown_event, self.wandb_queue), daemon=False
        )
        ps.start()

    def shutdown(self) -> None:
        logger.info("set shutdown event")
        self.shutdown_event.set()

    def safe_entrypoint(self, shutdown_event: multiprocessing.synchronize.Event, wandb_queue) -> None:  # type: ignore[no-untyped-def]
        # We don't wrap this in a loop because it really shouldn't fail flakily.
        # This is just for debugging a fatal crash.
        try:
            configure_parent_logging()
            self.entrypoint(shutdown_event, wandb_queue)
        except Exception as e:
            logger.warning("exception in AsyncGamePlayer process")
            logger.warning(e)
            logger.warning(traceback.format_exc())
            # This basically just gets eaten
            raise
        finally:
            logger.info("closing queue")
            self.wandb_queue.close()
            logger.info("queue closed")

    def entrypoint(self, shutdown_event: multiprocessing.synchronize.Event, wandb_queue) -> None:  # type: ignore[no-untyped-def]
        setup_new_process()
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        seed_and_run_deterministically_if_enabled()

        params = attr.evolve(
            self.params,
            env_params=attr.evolve(
                self.params.env_params,
                env_index=self.params.env_params.env_index + self.rollout_manager_id * self.params.num_workers,
            ),
        )
        # Spread among the inference gpus evenly
        rollout_device = torch.device(
            f"cuda:{self.params.inference_gpus[self.rollout_manager_id % len(self.params.inference_gpus)]}"
        )
        storage = EpisodeStorage(
            self.params,
            rollout_dir=self.train_rollout_dir,
            wandb_queue=wandb_queue,
            step_data_type=StepData,
            num_workers=params.num_workers,
        )
        algorithm_cls = get_algorithm_cls(self.params)
        model = algorithm_cls(params, self.obs_space, self.action_space)

        if self.params.resume_from:
            checkpoint_path = get_checkpoint_file(self.params.resume_from)
            model.load_state_dict(torch.load(checkpoint_path, map_location=self.params.train_device))
            logger.info("RESUMED MODEL FROM CHECKPOINT")

        # You cannot have initialized cuda in the parent process before forking for this to work.
        model = model.to(rollout_device)
        while True:
            if shutdown_event.is_set():
                break
            player = RolloutManager(
                params=self.params,
                num_workers=self.params.num_workers,
                is_multiprocessing=self.params.multiprocessing,
                storage=storage,
                obs_space=self.obs_space,
                model=model,
                rollout_device=rollout_device,
                multiprocessing_context=self.multiprocessing_context,
            )
            while True:
                if shutdown_event.is_set():
                    # TODO: we should really pass this through to the Manager - so it stops immediately,
                    # versus on the next iteration (which could be a bit).
                    logger.info("shutting down player")
                    player.shutdown()
                    logger.info("shutting down storage")
                    storage.shutdown()
                    logger.info("done with shutdown")
                    break
                try:
                    start = time.time()
                    ModelStorage.load_model(model, rollout_device)
                    # TODO: convert to a non-fixed step count
                    player.run_rollout(
                        num_steps=self.params.num_steps,
                        exploration_mode="explore",
                    )
                    rollout_fps = (self.params.num_workers * self.params.num_steps) / (time.time() - start)
                    logger.info(
                        f"player {self.rollout_manager_id}: env steps: {self.env_step_counter.value}, fps:{rollout_fps}"
                    )
                    self.env_step_counter.value += self.params.num_workers * self.params.num_steps
                    wandb_queue.put(("scalar", "env_step", self.env_step_counter.value))
                except Exception as e:
                    if self.params.debug:
                        raise
                    logger.warning("caught exception in async worker, will try to reset.")
                    logger.warning(e)
                    logger.warning(traceback.format_exc())
                    # capture_exception(e)
                    try:
                        player.shutdown()
                        storage.shutdown()
                    except Exception as e2:
                        logger.warning(e2)
                        logger.warning(traceback.format_exc())
                    break


class RolloutManager:
    """A manager class for running multiple game-playing processes."""

    def __init__(
        self,
        params: Params,
        num_workers: int,
        is_multiprocessing: bool,
        storage: TrajectoryStorage,
        obs_space: gym.spaces.Dict,
        model: Algorithm,
        rollout_device: torch.device,
        multiprocessing_context: BaseContext,
    ) -> None:
        self.multiprocessing = is_multiprocessing
        self.num_workers = num_workers
        self.params = params
        self.model = model
        self.rollout_device = rollout_device
        self.multiprocessing_context = multiprocessing_context

        params = attr.evolve(params, env_params=attr.evolve(params.env_params, env_count=num_workers))

        self.step_storage = self.setup_step_storage()

        # Start game-playing processes
        self.workers = [
            EnvironmentContainer(i, params, is_multiprocessing, multiprocessing_context, self.step_storage)
            for i in range(num_workers)
        ]

        self.storage = storage
        self.observation_space = obs_space

    def setup_step_storage(self):
        # Note the use of Tensor.share_memory_() for the components that are shared with the workers

        # the ordering here is (reward, done, obs (next_obs)), and then from next_obs we compute (action, value, etc)
        # so depending on which substep you're on, the ordering of the two groups will differ.
        # but key is that the observation is the one following the reward/done.
        return {
            "terminal_obs": create_observation_storage(
                self.params.observation_space, batch_shape=(self.num_workers,), use_shared_memory=True
            ),
            "obs_for_inference": create_observation_storage(
                self.params.observation_space, batch_shape=(self.num_workers,), use_shared_memory=True
            ),
            "obs_to_store": create_observation_storage(
                self.params.observation_space, batch_shape=(self.num_workers,), use_shared_memory=True
            ),
            "action": create_action_storage(
                self.params.action_space, batch_shape=(self.num_workers,), use_shared_memory=True
            ),
            "reward": torch.zeros((self.num_workers,), dtype=torch.float32).share_memory_(),
            "done": torch.zeros((self.num_workers,), dtype=torch.bool).share_memory_(),
            "is_terminal": torch.zeros((self.num_workers,), dtype=torch.bool).share_memory_(),
            "ready_for_new_step": torch.ones((self.num_workers,), dtype=torch.bool).share_memory_(),
            "close": torch.zeros((self.num_workers,), dtype=torch.bool).share_memory_(),
            "info": {
                k: torch.zeros((self.num_workers,), dtype=torch.float32).share_memory_()
                for k in self.params.env_params.info_fields
            },
        }

    def run_rollout(
        self,
        num_steps: Optional[int] = None,
        num_episodes: Optional[int] = None,
        exploration_mode: str = "explore",
    ) -> None:
        """Set num_steps and/or num_episodes to run only that many steps/episodes,
        otherwise if left as None there will be no limit. Meets both criteria if both are set.
        """
        self.model.eval()
        fragment_timestep = torch.zeros((self.num_workers,), dtype=torch.int32)

        # Note that we don't reset step_storage between rollouts.
        # This is intentional - "obs_for_inference" is used for generating the first action,
        # and dones need to be carried over too for algorithms that use them in inference.
        # And the rest could be zeroed but isn't for performance reasons (possibly should be for aid in debugging tho).

        for i, worker in enumerate(self.workers):
            worker.remaining_steps = num_steps if num_steps is not None else 0
            worker.remaining_episodes = num_episodes if num_episodes is not None else 0
            worker.ready_for_new_step = True
            worker.completed_rollout = False

        while True:
            if self.num_active_workers == 0:
                break

            # run the model
            inference_mask = self.ready_for_new_step
            inference_timesteps = fragment_timestep
            step_actions, algorithm_extra_info = self.run_inference(exploration_mode)
            masked_copy_structure(self.step_storage["action"], step_actions, inference_mask)
            self.request_step()

            # send the inference data to storage. doing it here hides the time behind the step time.
            self.storage.partial_batch_update(
                "action", step_actions, mask=inference_mask, timesteps=inference_timesteps
            )
            for k, v in algorithm_extra_info.items():
                self.storage.partial_batch_update(k, v, mask=inference_mask, timesteps=inference_timesteps)

            step_time = time.time()
            threshold_time = None
            got_data = torch.zeros((self.num_workers,), dtype=torch.bool)
            while True:
                num_active_workers = self.num_active_workers
                ready_for_new_step = self.ready_for_new_step
                if num_active_workers == 0:
                    break
                if self.params.allow_partial_batches:
                    needed_percent = max(0.5, min(0.6, 1 - (1 / num_active_workers)))
                else:
                    needed_percent = 1
                if ready_for_new_step.sum() >= num_active_workers:
                    break
                if ready_for_new_step.sum() / num_active_workers + 1e-5 >= needed_percent:
                    if threshold_time is None:
                        threshold_time = time.time()
                    if (threshold_time - step_time) * 1.25 < (time.time() - step_time):
                        break

                # Receive step data from workers
                for i, worker in enumerate(self.workers):
                    if worker.completed_rollout or worker.ready_for_new_step or not worker.is_result_ready():
                        continue

                    # once we run this, the worker's data is ready to access
                    self.get_worker_data(worker, i)
                    got_data[i] = True

            # TODO: move this to when we're waiting on the stepping to hide the time
            for k in ["done", "reward", "is_terminal", "info"]:
                self.storage.partial_batch_update(k, self.step_storage[k], mask=got_data, timesteps=fragment_timestep)
            self.storage.partial_batch_update(
                "observation", self.step_storage["obs_to_store"], mask=got_data, timesteps=fragment_timestep
            )
            self.storage.mark_step_finished(got_data)
            fragment_timestep[got_data] += 1

        # We store the final observation too, so that training can use it for value bootstrapping
        # In the case of a reset, this will be from the next episode, and GAE should stop bootstrapping
        # in that case. It would be improper to give the terminal obs, since that is never otherwise provided
        # in on-policy training. And the normal `obs_to_store` is still from the previous step.
        # This will be ignored (from the flag is_extra_observation) in storage methods that don't use it.
        self.storage.partial_batch_update(
            "observation",
            self.step_storage["obs_for_inference"],
            mask=torch.ones((self.num_workers,), dtype=torch.bool),
            timesteps=torch.ones((self.num_workers,), dtype=torch.int32) * -1,
            is_extra_observation=True,
        )

    def request_step(self) -> None:
        for i, worker in enumerate(self.workers):
            if worker.ready_for_new_step and not worker.completed_rollout:
                worker.send_step()

    def run_inference(self, exploration_mode: str) -> Tuple[ActionBatch, dict]:
        with torch.no_grad():
            torch_obs = {
                k: v.to(device=self.rollout_device, non_blocking=True)
                for k, v in self.step_storage["obs_for_inference"].items()
            }
            torch_dones = self.step_storage["done"].to(device=self.rollout_device, non_blocking=True)
            torch_obs = postprocess_uint8_to_float(torch_obs, center=self.params.center_observations)
            # we pass ready_for_new_step to allow the model to not update state for non-ready workers.
            # But we expect to get the num_workers as the returned batch size (some are ignored).
            step_actions, to_store = self.model.rollout_step(
                torch_obs, torch_dones, self.ready_for_new_step, exploration_mode=exploration_mode
            )
        return step_actions, to_store

    def get_worker_data(
        self,
        worker: "EnvironmentContainer",
        i: int,
    ) -> None:
        done = worker.get_step_data()
        truncated = done and not self.step_storage["is_terminal"][i]
        if self.params.time_limit_bootstrapping and truncated:
            terminal_observation = map_structure(lambda x: x[i], self.step_storage["terminal_obs"])
            additional_reward = self.time_limit_bootstrapping(terminal_observation)
            self.step_storage["reward"][i] = self.step_storage["reward"][i] + additional_reward

    @property
    def ready_for_new_step(self) -> Tensor:
        return torch.tensor([worker.ready_for_new_step for worker in self.workers], dtype=torch.bool)

    @property
    def num_active_workers(self) -> int:
        return sum([worker.completed_rollout is False for worker in self.workers])

    def time_limit_bootstrapping(self, terminal_obs: dict[str, Tensor]) -> float:
        # NOTE: this won't work (as written) for recurrent algorithms!!
        # value bootstrapping for envs with artificial time limits imposed by a gym TimeLimit wrapper
        # see https://arxiv.org/pdf/1712.00378.pdf for context
        # this does get wrapper obs transforms applied
        terminal_obs_torch = {k: v.to(self.rollout_device).unsqueeze(0) for k, v in terminal_obs.items()}
        terminal_obs_torch = postprocess_uint8_to_float(terminal_obs_torch, center=self.params.center_observations)
        with torch.no_grad():
            terminal_value, _ = self.model(terminal_obs_torch)
        return self.params.discount * terminal_value[0].detach().cpu().item()  # type: ignore

    def shutdown(self) -> None:
        [worker.close() for worker in self.workers]


class EnvironmentContainer:
    """Abstracts away the details of interacting with a worker."""

    def __init__(
        self,
        i: int,
        params: Params,
        is_multiprocessing: bool,
        multiprocessing_context: BaseContext,
        step_storage: dict[str, Any],
    ) -> None:
        self.params = params
        self.i = i
        self.is_multiprocessing = is_multiprocessing
        env_params = attr.evolve(params.env_params, env_index=params.env_params.env_index + i)
        if is_multiprocessing:
            parent_conn, child_conn = multiprocessing_context.Pipe()
            worker = EnvironmentContainerProcess(i, child_conn, env_params, step_storage, params.obs_first)
            # Can't have daemon=True because the godot env spawns subprocesses
            self.ps = multiprocessing_context.Process(target=worker.run, args=(), daemon=False)
            # self.ps = threading.Thread(target=worker.run, args=())
            self.ps.start()
            self.parent_conn = parent_conn
        else:
            self.worker = EnvironmentContainerProcess(i, None, env_params, step_storage, params.obs_first)

        self.ready_for_new_step: bool = True
        self.completed_rollout: bool = False
        # maybe it would be more intuitive to count up instead of down here?
        self.remaining_steps: int = 0
        self.remaining_episodes: int = 0
        self.step_storage = step_storage

        assert params.observation_space is not None

    def get_step_data(self):
        # The side-effect of this function returning is that we are guaranteed the obs/reward/etc memory
        # arrays have been written by the worker.
        done = self.step_storage["done"][self.i]

        self.ready_for_new_step = True
        self.remaining_steps = max(0, self.remaining_steps - 1)
        if done:
            self.remaining_episodes = max(0, self.remaining_episodes - 1)
        if self.remaining_steps == 0 and self.remaining_episodes == 0:
            self.ready_for_new_step = False
            self.completed_rollout = True

        return done

    def send_step(self) -> None:
        if self.is_multiprocessing:
            self.step_storage["ready_for_new_step"][self.i] = False
        else:
            self.worker.step()
        self.ready_for_new_step = False

    def is_result_ready(self) -> bool:
        if self.is_multiprocessing:
            return self.step_storage["ready_for_new_step"][self.i].item() is True
        else:
            return True

    def close(self) -> None:
        if self.is_multiprocessing:
            try:
                self.step_storage["close"][self.i] = True
            except BrokenPipeError:
                logger.warning("tried to close EnvironmentContainerProcess but pipe was already broken")
            except EOFError:
                logger.warning("tried to close EnvironmentContainerProcess but got EOF error.")
        else:
            self.worker.shutdown()


class EnvironmentContainerProcess:
    """A worker for running an environment, intended to be run on a separate
    process."""

    def __init__(  # type: ignore
        self,
        index: int,
        pipe,  # multiprocessing.Pipe (impossible to type afaik)
        env_params: EnvironmentParams,
        step_storage: dict[str, Any],
        obs_first: bool,
    ) -> None:
        self.worker_index = index
        self.pipe = pipe
        self.episode_steps: int = 0
        self.episode_rewards: float = 0

        self.env_params = env_params
        self.env: Optional[gym.Env] = None

        self.step_storage = step_storage
        self.obs_first = obs_first

    def lazy_init_env(self) -> None:
        if self.env is None:
            setup_sentry()
            self.env = build_env(env_params=self.env_params)
            self.next_obs = self.env.reset()
            self.next_obs = {k: v * 0 for k, v in self.next_obs.items()}

    def shutdown(self) -> None:
        if self.env:
            try:
                self.env.close()
            except TimeoutError as e:
                capture_exception(e)
                logger.warning(f"env did not close cleanly due to a TimeoutError: {e}")
                logger.warning(traceback.format_exc())
            except Exception as e:
                capture_exception(e)
                logger.warning(f"worker failed to shutdown with error {e}")
                logger.warning(traceback.format_exc())

    def run(self):
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        configure_parent_logging()
        setup_new_process()
        try:
            while True:
                if self.step_storage["ready_for_new_step"][self.worker_index] == 0:
                    self.step()
                elif self.step_storage["close"][self.worker_index]:
                    break
                else:
                    time.sleep(2e-4)  # 200us
        except KeyboardInterrupt:
            pass
        except Exception as e:
            capture_exception(e)
            raise
        finally:
            self.shutdown()

    def step(self):
        """Perform a single step of the environment."""
        self.lazy_init_env()
        assert self.env is not None
        action = map_structure(lambda x: x[self.worker_index], self.step_storage["action"])
        next_obs, reward, done, info = self.env.step(action)  # type: ignore

        self.episode_rewards += reward

        if done:
            terminal_obs = next_obs
            # We need to store this for value bootstrapping, even if it's not used for inference or training
            for k, v in next_obs.items():
                self.step_storage["terminal_obs"][k][self.worker_index] = v

            next_obs = self.env.reset()
            self.episode_steps = 0
            self.episode_rewards = 0

        self.episode_steps += self.env_params.action_repeat

        is_terminal = done and not info.get("TimeLimit.truncated", False)
        assert isinstance(reward, (int, float))

        if self.obs_first:
            # Since `next_obs` comes after the reward/done, we need to use the obs from last step
            obs_to_store = self.next_obs
            # if the ep was done, next_obs is from the reset() call,
            # which will be the right thing to store on the next step.
            # Note: in this case, we never store (or use at all) the terminal observation. This is correct.
            self.next_obs = next_obs
        else:
            if done:
                # We need to handle this case differently. next_obs is actually from the next ep (from reset());
                # it wouldn't be good to store that as part of this one.
                # Instead we store the "terminal observation" which is the one received along with the done signal.
                # We handle both the time_limit and no_time_limit case the same.
                # NOTE: Interestingly, this means the agent never sees the first observation of an episode
                # (ie the one that comes from reset()) in training (but it does in inference).
                obs_to_store = terminal_obs  # type: ignore
            else:
                obs_to_store = next_obs

        # inference always uses next_obs; after a done, this is the obs received from reset().
        for k, v in next_obs.items():
            self.step_storage["obs_for_inference"][k][self.worker_index] = v
        # but we store the observation we selected above
        for k, v in obs_to_store.items():
            self.step_storage["obs_to_store"][k][self.worker_index] = v

        self.step_storage["reward"][self.worker_index] = reward
        self.step_storage["done"][self.worker_index] = done
        self.step_storage["is_terminal"][self.worker_index] = is_terminal

        self.step_storage["ready_for_new_step"][self.worker_index] = True

        for k in self.env_params.info_fields:
            self.step_storage["info"][k][self.worker_index] = info.get(k, 0)

        assert "terminal_observation" not in info  # we need to keep the infos small to keep this fast.
        return done, info



# %%
import json
import multiprocessing as mp
import os
import tarfile
import traceback
from datetime import datetime
from pathlib import Path
from random import shuffle
from typing import Dict
from typing import Optional
from typing import Tuple

import nest
import sentry_sdk
import torch

from avalon.agent.evaluation import EVAL_TEMP_PATH
from avalon.agent.evaluation import EvaluationGodotEnv
from avalon.agent.evaluation import get_wandb_result_key
from avalon.agent.evaluation import get_world_folders
from avalon.agent.evaluation import world_id_from_world_folder_name
from avalon.agent.torchbeast.avalon_helpers import force_cudnn_initialization
from avalon.agent.torchbeast.avalon_helpers import wrap_evaluation_godot_env
from avalon.agent.torchbeast.core.environment import Environment
from avalon.agent.torchbeast.polybeast_learner import Net
from avalon.common.error_utils import setup_sentry
from avalon.common.log_utils import configure_parent_logging
from avalon.common.log_utils import logger
from avalon.common.wandb_utils import load_checkpoint_from_wandb_run
from avalon.contrib.s3_utils import SimpleS3Client


@torch.no_grad()
def do_eval_rollouts(model: Net, env: Environment):
    force_cudnn_initialization()
    # DO ROLLOUTS
    observation = env.initial()
    core_state = model.initial_state(batch_size=1)
    core_state = nest.map(lambda t: t.to("cuda:0"), core_state)
    scores_by_world_index: Dict[int, float] = {}
    frame_count = 1
    while True:
        (flat_action, policy_logits, baseline), core_state = model(observation, core_state)
        observation = env.step(flat_action)
        frame_count += 1
        world_index = observation["info"]["world_index"]

        if world_index in scores_by_world_index:
            break
        if observation["done"].item():
            info = observation["info"]
            logger.info(
                f"Episode {world_index}, {info['task']} ({info['difficulty']}) score: {info['score']:.3f} in {frame_count} frames"
            )
            scores_by_world_index[world_index] = observation["info"]["score"]
            frame_count = 1
    return scores_by_world_index


def run_checkpoint_on_world_ids(
    checkpoint_path: str, eval_world_dir: str, world_ids: Tuple[int, ...], gpu_id: int = 0
):
    device = torch.device(f"cuda:{gpu_id}")
    env = wrap_evaluation_godot_env(EvaluationGodotEnv(eval_world_dir=eval_world_dir, gpu_id=gpu_id))
    env.gym_env.set_eval_world_ids(world_ids)
    model = Net(use_lstm=True).to(device=device)
    checkpoint_states = torch.load(checkpoint_path, map_location=device)
    model.load_state_dict(checkpoint_states["model_state_dict"])
    result = do_eval_rollouts(model, env)
    return result


def run_evaluation(worlds_path: Path, data_key: Optional[str], wandb_run: str, checkpoint_filename: str) -> None:

    s3_client = SimpleS3Client()
    result_key = get_wandb_result_key(wandb_run, checkpoint_filename, data_key)

    setup_sentry(percent_of_traces_to_capture=1.0)

    if data_key is None:
        assert worlds_path.is_dir()
    else:
        logger.info(f"Downloading data from {data_key} to {worlds_path}")
        os.makedirs(worlds_path, exist_ok=True)
        os.makedirs(EVAL_TEMP_PATH, exist_ok=True)
        s3_client.download_to_file(data_key, Path(EVAL_TEMP_PATH) / data_key)
        with tarfile.open(Path(EVAL_TEMP_PATH) / data_key, "r:gz") as f:
            f.extractall(worlds_path)

    logger.info(f"Downloading checkpoint {checkpoint_filename} from wandb run {wandb_run}")
    checkpoint_path = load_checkpoint_from_wandb_run(wandb_run, checkpoint_filename)

    try:
        previous_result = json.loads(s3_client.load(result_key))
        all_results = previous_result["all_results"]
        logger.info(f"Skipping {len(all_results)} already completed worlds")
    except KeyError:
        all_results = {}
    all_errors = []

    num_processes = 5
    world_id_list = [world_id_from_world_folder_name(x) for x in get_world_folders(str(worlds_path))]
    world_id_list = [x for x in world_id_list if str(x) not in all_results]
    shuffle(world_id_list)
    num_worlds_per_process = len(world_id_list) // num_processes + min(1, len(world_id_list) % num_processes)

    process_world_ids = [
        tuple(world_id_list[i : i + num_worlds_per_process])
        for i in range(0, len(world_id_list), num_worlds_per_process)
    ]
    logger.info(f"Splitting workload into {[len(x) for x in process_world_ids]}")

    def on_done(result) -> None:
        logger.info(f"Finished generating {result}")
        all_results.update(result)

    def on_error(error: BaseException) -> None:
        logger.error(f"Evaluation failed! {error}")
        traceback.print_exc()
        sentry_sdk.capture_exception(error)

    start_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

    ctx = mp.get_context("spawn")
    with ctx.Pool(processes=num_processes, initializer=configure_parent_logging) as worker_pool:
        requests = []
        for world_ids in process_world_ids:
            request = worker_pool.apply_async(
                run_checkpoint_on_world_ids,
                kwds={"checkpoint_path": checkpoint_path, "eval_world_dir": str(worlds_path), "world_ids": world_ids},
                callback=on_done,
                error_callback=on_error,
            )
            requests.append(request)
        worker_pool.close()
        worker_pool.join()
    end_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

    record_data = {
        "wandb_run": wandb_run,
        "baseline": "IMPALA",
        "checkpoint_filename": checkpoint_filename,
        "data_key": data_key,
        "eval_world_dir": str(worlds_path),
        "start_time": start_time,
        "end_time": end_time,
        "all_results": all_results,
    }

    logger.info(f"Saving result to '{result_key}'")
    s3_client.save(result_key, json.dumps(record_data))


if __name__ == "__main__":

    wandb_run = "sourceress/abe__torchbeast/38z8lhh9"
    checkpoint_filename = "model_step_50003200.tar"
    worlds_path = Path("/tmp/avalon_worlds/b21aeff3-59c5-49a8-9c19-f3feceea1b8a")
    data_key = "avalon_worlds__b21aeff3-59c5-49a8-9c19-f3feceea1b8a.tar.gz"
    run_evaluation(
        worlds_path=worlds_path, data_key=data_key, wandb_run=wandb_run, checkpoint_filename=checkpoint_filename
    )



import gzip
import json
import os
import shutil
import struct
from collections import defaultdict
from io import BufferedReader
from math import prod
from pathlib import Path
from typing import Any
from typing import BinaryIO
from typing import Dict
from typing import List
from typing import Literal
from typing import NamedTuple
from typing import Optional
from typing import OrderedDict as OrderedDictType
from typing import Sequence
from typing import Tuple
from typing import Type
from typing import Union
from typing import cast

import attr
import numpy as np
import numpy.typing as npt
from loguru import logger
from numpy import typing as npt

from avalon.common.errors import SwitchError
from avalon.datagen.env_helper import create_env
from avalon.datagen.godot_env.action_log import GodotEnvActionLog
from avalon.datagen.godot_env.actions import AttrsAction
from avalon.datagen.godot_env.actions import VRAction
from avalon.datagen.godot_env.actions import _from_bytes
from avalon.datagen.godot_env.actions import _to_bytes
from avalon.datagen.godot_env.observations import FAKE_TYPE_IMAGE
from avalon.datagen.godot_env.observations import NP_DTYPE_MAP
from avalon.datagen.godot_env.observations import AvalonObservation
from avalon.datagen.godot_env.observations import FeatureSpecDict
from avalon.datagen.godot_generated_types import ACTION_MESSAGE
from avalon.datagen.godot_generated_types import CLOSE_MESSAGE
from avalon.datagen.godot_generated_types import HUMAN_INPUT_MESSAGE
from avalon.datagen.godot_generated_types import QUERY_AVAILABLE_FEATURES_MESSAGE
from avalon.datagen.godot_generated_types import RESET_MESSAGE
from avalon.datagen.godot_generated_types import SEED_MESSAGE
from avalon.datagen.godot_generated_types import SELECT_FEATURES_MESSAGE
from avalon.datagen.godot_generated_types import AvalonSimSpec
from avalon.datagen.godot_generated_types import MouseKeyboardAgentPlayerSpec
from avalon.datagen.godot_generated_types import MouseKeyboardHumanPlayerSpec
from avalon.datagen.godot_generated_types import RecordingOptionsSpec
from avalon.datagen.godot_generated_types import VRAgentPlayerSpec
from avalon.datagen.godot_generated_types import VRHumanPlayerSpec


class PlaybackError(NamedTuple):
    frame: int
    key: str
    recorded_value: npt.NDArray[np.float32]
    playback_value: npt.NDArray[np.float32]
    max_error: float


class PlaybackResult(NamedTuple):
    is_error: bool
    errors: Dict[int, List[PlaybackError]]
    recorded_observations: List[AvalonObservation]
    playback_observations: List[AvalonObservation]
    human_inputs: List[AttrsAction]
    actions: List[AttrsAction]


SKIP_KEYS = ["rgb", "depth", "rgbd", "isometric_rgbd", "top_down_rgbd"]
DISCRETE_KEYS = [
    "left_hand_thing_colliding_with_hand",
    "left_hand_held_thing",
    "right_hand_thing_colliding_with_hand",
    "right_hand_held_thing",
    "nearest_food_id",
]


def _write_message(fp: BinaryIO, message_type: int, message: bytes) -> None:
    fp.write(message_type.to_bytes(1, byteorder="little", signed=False))
    fp.write(message)


def get_replay_log_from_human_recording(
    new_action_record_path: Path,
    action_record_path: Path,
    metadata_path: Path,
    worlds_path: Path,
    selected_features: FeatureSpecDict,
    action_type: Type[AttrsAction],
) -> GodotEnvActionLog:
    with open(metadata_path, "r") as metadata_file:
        metadata = json.load(metadata_file)

    episode_seed = metadata["episode_seed"]
    world_id = metadata["world_id"]
    world_path = get_world_path_from_world_id(root_path=worlds_path, world_id=world_id)

    with open(action_record_path, "rb") as action_file:
        content = action_file.read()

    Path(new_action_record_path).unlink(missing_ok=True)

    # need to manually add some header messages to our replay log
    with open(new_action_record_path, "wb") as new_action_file:
        _write_message(new_action_file, QUERY_AVAILABLE_FEATURES_MESSAGE, bytes())

        size_doubleword = (len(selected_features)).to_bytes(4, byteorder="little", signed=False)
        feature_names_bytes = ("\n".join(selected_features.keys()) + "\n").encode("UTF-8")
        _write_message(new_action_file, SELECT_FEATURES_MESSAGE, size_doubleword + feature_names_bytes)

        seed_message_bytes = episode_seed.to_bytes(8, byteorder="little", signed=True)
        _write_message(new_action_file, SEED_MESSAGE, seed_message_bytes)

        reset_message_bytes = (
            action_type.get_null_action().to_bytes()
            + episode_seed.to_bytes(8, byteorder="little", signed=True)
            + (world_path + "\n").encode("UTF-8")
            + _to_bytes(float, 1.0)
        )
        _write_message(new_action_file, RESET_MESSAGE, reset_message_bytes)

        new_action_file.write(content)

        _write_message(new_action_file, CLOSE_MESSAGE, bytes())

    return GodotEnvActionLog.parse(str(new_action_record_path), action_type)


def get_observations_from_human_recording(
    observations_path: Path, selected_features: OrderedDictType[str, Tuple[int, Tuple[int, ...]]]
) -> List[AvalonObservation]:
    file_size = os.stat(observations_path).st_size
    observations = []
    with open(observations_path, "rb") as record_log:
        while record_log.tell() < file_size:
            feature_data = {}
            for feature_name, (data_type, dims) in selected_features.items():
                # TODO make a constant / is this still a problem?
                if feature_name in {"depth", "rgb", "rgbd"}:
                    continue
                size = prod(dims)
                if data_type != FAKE_TYPE_IMAGE:
                    size = size * 4
                feature_data[feature_name] = _read_shape(record_log, dims, size, dtype=NP_DTYPE_MAP[data_type])

            cleaned_feature_data = {
                feature.name: feature_data.get(feature.name, np.array([]))
                for feature in attr.fields(AvalonObservation)
            }
            observations.append(AvalonObservation(**cleaned_feature_data))
    return observations


def _read_shape(
    record_log: BufferedReader,
    shape: Tuple[int, ...],
    size: Optional[int] = None,
    dtype: npt.DTypeLike = np.uint8,
) -> npt.NDArray:
    byte_buffer = record_log.read(size if size is not None else prod(shape))

    return np.ndarray(shape=shape, dtype=dtype, buffer=byte_buffer)


def get_world_path_from_world_id(root_path: Path, world_id: str) -> str:
    return str(root_path / world_id / "main.tscn")


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class MouseKeyboardHumanInputAction(AttrsAction):
    is_move_forward_pressed: float
    is_move_backward_pressed: float
    is_move_left_pressed: float
    is_move_right_pressed: float
    is_jump_pressed: float
    is_eat_pressed: float
    is_grab_pressed: float
    is_throw_pressed: float
    is_crouch_pressed: float
    is_wheel_up_just_released: float
    is_wheel_down_just_released: float
    is_mouse_mode_toggled: float
    is_active_hand_toggled: float


def _to_bytes_with_double(value_type: Type, value: Any) -> bytes:
    if value_type is float:
        return struct.pack("<d", value)
    if value_type is int:
        return struct.pack("<i", value)
    raise SwitchError(f"{value_type} (value={value})")


def _from_bytes_with_double(value_type: Type, value_bytes: bytes) -> Tuple[Union[int, float], bytes]:
    byte_format = ""
    if value_type is float:
        byte_format = "<d"
    elif value_type is int:
        byte_format = "<i"
    else:
        raise SwitchError(f"{value_type} should be either float or int")

    size = struct.calcsize(byte_format)
    bytes_to_consume, bytes_remaining = value_bytes[:size], value_bytes[size:]
    return struct.unpack(byte_format, bytes_to_consume)[0], bytes_remaining


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class VRHumanInputAction(AttrsAction):
    human_height: float
    arvr_origin_position_x: float
    arvr_origin_position_y: float
    arvr_origin_position_z: float
    arvr_origin_basis_x_x: float
    arvr_origin_basis_x_y: float
    arvr_origin_basis_x_z: float
    arvr_origin_basis_y_x: float
    arvr_origin_basis_y_y: float
    arvr_origin_basis_y_z: float
    arvr_origin_basis_z_x: float
    arvr_origin_basis_z_y: float
    arvr_origin_basis_z_z: float
    arvr_camera_position_x: float
    arvr_camera_position_y: float
    arvr_camera_position_z: float
    arvr_camera_basis_x_x: float
    arvr_camera_basis_x_y: float
    arvr_camera_basis_x_z: float
    arvr_camera_basis_y_x: float
    arvr_camera_basis_y_y: float
    arvr_camera_basis_y_z: float
    arvr_camera_basis_z_x: float
    arvr_camera_basis_z_y: float
    arvr_camera_basis_z_z: float
    arvr_left_hand_position_x: float
    arvr_left_hand_position_y: float
    arvr_left_hand_position_z: float
    arvr_left_hand_basis_x_x: float
    arvr_left_hand_basis_x_y: float
    arvr_left_hand_basis_x_z: float
    arvr_left_hand_basis_y_x: float
    arvr_left_hand_basis_y_y: float
    arvr_left_hand_basis_y_z: float
    arvr_left_hand_basis_z_x: float
    arvr_left_hand_basis_z_y: float
    arvr_left_hand_basis_z_z: float
    arvr_right_hand_position_x: float
    arvr_right_hand_position_y: float
    arvr_right_hand_position_z: float
    arvr_right_hand_basis_x_x: float
    arvr_right_hand_basis_x_y: float
    arvr_right_hand_basis_x_z: float
    arvr_right_hand_basis_y_x: float
    arvr_right_hand_basis_y_y: float
    arvr_right_hand_basis_y_z: float
    arvr_right_hand_basis_z_x: float
    arvr_right_hand_basis_z_y: float
    arvr_right_hand_basis_z_z: float
    look_stick_x: float
    look_stick_y: float
    strafe_stick_x: float
    strafe_stick_y: float
    is_left_hand_grab_pressed: float
    is_right_hand_grab_pressed: float
    is_jump_pressed: float

    @classmethod
    def from_bytes(cls: Type["VRHumanInputAction"], action_bytes: bytes) -> "VRHumanInputAction":
        fields: Dict[str, Any] = {}
        _size, remaining_bytes = _from_bytes_with_double(int, action_bytes)
        for field in attr.fields(cls):
            fields[field.name], remaining_bytes = _from_bytes_with_double(cast(Type, field.type), remaining_bytes)
        return cls(**fields)

    def to_bytes(self) -> bytes:
        action_bytes = b"".join(
            _to_bytes_with_double(
                cast(Type, field.type),
                getattr(self, field.name),
            )
            for field in attr.fields(self.__class__)
        )
        return _to_bytes_with_double(int, len(action_bytes)) + action_bytes


def parse_human_input(
    path: Path, human_input_type: Union[Type[VRHumanInputAction], Type[MouseKeyboardHumanInputAction]]
) -> Sequence[AttrsAction]:
    human_inputs = []
    with open(path, "rb") as f:
        while message_bytes := f.read(1):
            message = int.from_bytes(message_bytes, byteorder="little", signed=False)
            assert message == HUMAN_INPUT_MESSAGE
            size_bytes = f.read(4)
            size, _ = _from_bytes(int, size_bytes)
            human_input_bytes = size_bytes + f.read(cast(int, size))
            human_inputs.append(human_input_type.from_bytes(human_input_bytes))

    return human_inputs


def get_oculus_config_for_human_playback() -> AvalonSimSpec:
    # NOTE: make sure you get the type of player correct in config.json
    config = AvalonSimSpec.from_dict(json.load(open("datagen/godot/android/config.json", "r")))

    # update player to be agent player
    with config.mutable_clone() as config:
        # update recording options to make sense for the agent
        config.recording_options = RecordingOptionsSpec(
            user_id=None,
            apk_version=None,
            recorder_host=None,
            recorder_port=None,
            is_teleporter_enabled=False,
            resolution_x=64,
            resolution_y=64,
            is_recording_human_actions=False,
            is_remote_recording_enabled=False,
            is_adding_debugging_views=False,
            is_debugging_output_requested=False,
        )
        if isinstance(config.player, MouseKeyboardHumanPlayerSpec):
            config.player = MouseKeyboardAgentPlayerSpec.from_dict(config.player.to_dict())
        elif isinstance(config.player, VRHumanPlayerSpec):
            player_dict = config.player.to_dict()
            player_dict["arm_length"] = 10_000
            # player_dict["total_energy_coefficient"] = 1e-4
            # make vr arm length basically infinity
            config.player = VRAgentPlayerSpec.from_dict(player_dict)
        else:
            raise SwitchError(config.player)

    return config


def get_oculus_playback_config(is_using_human_input: bool):
    # NOTE: make sure you get the type of player correct in config.json
    config = AvalonSimSpec.from_dict(json.load(open(Path(__file__).parent / "godot/android/config.json", "r")))

    # update player to be agent player
    with config.mutable_clone() as config:
        # update recording options to make sense for the agent
        config.recording_options = RecordingOptionsSpec(
            user_id=None,
            apk_version=None,
            recorder_host=None,
            recorder_port=None,
            is_teleporter_enabled=False,
            resolution_x=64,
            resolution_y=64,
            is_recording_human_actions=False,
            is_remote_recording_enabled=False,
            is_adding_debugging_views=False,
            is_debugging_output_requested=False,
        )
        if isinstance(config.player, MouseKeyboardHumanPlayerSpec):
            config.player = MouseKeyboardAgentPlayerSpec.from_dict(config.player.to_dict())
        elif isinstance(config.player, VRHumanPlayerSpec):
            player_dict = config.player.to_dict()
            player_dict["arm_length"] = 10_000
            player_dict["is_human_playback_enabled"] = is_using_human_input
            # make vr arm length basically infinity

            if is_using_human_input:
                config.player = VRHumanPlayerSpec.from_dict(player_dict)
            else:
                config.player = VRAgentPlayerSpec.from_dict(player_dict)
        else:
            raise SwitchError(config.player)
    return config


def validate_oculus_playback_recording(
    playback_path: Path,
    worlds_path: Path,
    threshold: float,
    is_using_human_input: bool,
) -> PlaybackResult:
    new_action_record_path = playback_path / "new_actions.out"
    action_record_path = playback_path / "actions.out"
    metadata_path = playback_path / "metadata.json"
    observations_path = playback_path / "observations.out"
    human_input_path = playback_path / "human_inputs.out"

    data_paths = [action_record_path, metadata_path, observations_path, human_input_path]

    for raw_path in data_paths:
        path = Path(f"{raw_path}.gz")
        if path.exists() and not raw_path.exists():
            with gzip.open(str(path), "rb") as f_in:
                with open(raw_path, "wb") as f_out:
                    shutil.copyfileobj(f_in, f_out)

    config = get_oculus_playback_config(is_using_human_input)
    assert config.player is not MouseKeyboardAgentPlayerSpec and config.player is not MouseKeyboardHumanPlayerSpec

    action_type: Union[Type[VRHumanInputAction], Type[VRAction]]
    if is_using_human_input:
        action_path = human_input_path
        action_type = VRHumanInputAction
    else:
        action_path = action_record_path
        action_type = VRAction

    env = create_env(config, action_type)
    selected_features = env.observation_context.selected_features
    # available_features = env.observation_context.available_features
    env.close()

    human_observations = get_observations_from_human_recording(
        observations_path=observations_path,
        selected_features=selected_features,
    )

    replay_log = get_replay_log_from_human_recording(
        new_action_record_path=new_action_record_path,
        action_record_path=action_path,
        metadata_path=metadata_path,
        worlds_path=worlds_path,
        selected_features=selected_features,
        action_type=action_type,
    )

    metadata = json.load(open(metadata_path))
    env = create_env(config, action_type)

    episode_id = replay_log.initial_episode_id
    world_id = metadata["world_id"]
    world_path = get_world_path_from_world_id(root_path=Path(worlds_path), world_id=world_id)
    env.seed_nicely(episode_id)

    actions = []
    observations = []
    for message in replay_log.messages:
        if message[0] in (ACTION_MESSAGE, HUMAN_INPUT_MESSAGE):
            # note: mypy cannot infer this type on its own
            message = cast(Tuple[Literal[3, 9], Union[VRAction, VRHumanInputAction]], message)
            action = message[1]
            obs, _ = env.act(action)
            actions.append(action)
            observations.append(obs)
        elif message[0] == RESET_MESSAGE:
            observations.append(env.reset_nicely_with_specific_world(episode_seed=episode_id, world_path=world_path))
        else:
            raise SwitchError(f"Invalid replay message {message}")

    human_inputs = [x for x in parse_human_input(human_input_path, VRHumanInputAction)]
    errors = defaultdict(list)
    for i, (agent_ob, player_ob) in enumerate(
        # note: when coming from human input, there's an extra observation
        zip(observations[len(observations) - len(human_observations) :], human_observations)
    ):
        agent_observation_dict = attr.asdict(agent_ob)
        player_observation_dict = attr.asdict(player_ob)
        for key in player_observation_dict:
            if key in SKIP_KEYS:
                continue
            agent_value = agent_observation_dict[key]
            player_value = player_observation_dict[key]
            if key in DISCRETE_KEYS:
                max_error = (
                    1 if (agent_value != -1 and player_value == -1 or agent_value == -1 and player_value != -1) else 0
                )
            else:
                if np.isinf(agent_value.max()) or np.isinf(player_value.max()):
                    max_error = 1 if np.any(agent_value != player_value) else 0
                else:
                    delta: npt.NDArray[np.float32] = agent_value - player_value
                    max_error = np.abs(delta).max().item()
            if max_error > threshold:
                # mark video for error
                errors[i].append(
                    PlaybackError(
                        frame=i,
                        key=key,
                        recorded_value=player_value,
                        playback_value=agent_value,
                        max_error=max_error,
                    )
                )

    return PlaybackResult(
        is_error=len(errors) > 0,
        errors=errors,
        recorded_observations=human_observations,
        playback_observations=observations,
        human_inputs=human_inputs,
        actions=cast(List[AttrsAction], actions),
    )


def validate_all_playback_recordings(
    worlds_root_path: Path, playback_root_path: Path, threshold: float, is_using_human_input: bool
) -> Dict[str, PlaybackResult]:
    results = {}
    for world_id_path in playback_root_path.iterdir():
        for playback_path in world_id_path.iterdir():
            key = f"{world_id_path.name}__{playback_path.name}"
            logger.debug(f"VALIDATING: {key}")
            results[key] = validate_oculus_playback_recording(
                playback_path, worlds_root_path, threshold, is_using_human_input
            )
            logger.debug("\n\n")
    return results



from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any
from typing import Callable
from typing import Generator
from typing import Iterable
from typing import Optional
from typing import Protocol
from typing import TypeVar
from typing import Union

import pytest
from _pytest.fixtures import Config
from _pytest.python import Function

from avalon.contrib.utils import TEMP_DIR
from avalon.contrib.utils import create_temp_file_path
from avalon.contrib.utils import temp_dir

if TYPE_CHECKING:
    # noinspection PyProtectedMember
    from _pytest.fixtures import _Scope

try:
    # noinspection PyProtectedMember
    from _pytest.fixtures import _FixtureFunction
except Exception:
    raise

T = TypeVar("T")


def fixture(
    fixture_function: Optional[_FixtureFunction] = None,
    *,
    scope: "Union[_Scope, Callable[[str, Config], _Scope]]" = "function",
    params: Optional[Iterable[object]] = None,
    autouse: bool = False,
    ids: Optional[
        Union[
            Iterable[Union[None, str, float, int, bool]],
            Callable[[Any], Optional[object]],
        ]
    ] = None,
) -> Any:
    def decorator(function: _FixtureFunction) -> Any:
        true_name = function.__name__[:-1]
        return pytest.fixture(function, name=true_name, scope=scope, params=params, autouse=autouse, ids=ids)

    if fixture_function is not None and callable(fixture_function):
        return decorator(fixture_function)

    return decorator


def use(*args: Callable[..., Any]) -> Any:
    true_names = [x.__name__[:-1] for x in args]
    return pytest.mark.usefixtures(*true_names)


def integration_test(function: Callable[..., None]) -> Any:
    return pytest.mark.integration_test(function)


def slow_integration_test(function: Callable[..., None]) -> Any:
    return pytest.mark.slow_integration_test(function)


class RequestFixture(Protocol[T]):
    """Yes, there is a class called FixtureRequest, but the types are quite bad for it"""

    node: Function
    param: T


@fixture
def temp_file_path_() -> Generator[Path, None, None]:
    with create_temp_file_path() as output:
        yield output


@fixture
def temp_path_() -> Generator[Path, None, None]:
    with temp_dir(TEMP_DIR) as output:
        yield output



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Union

import numpy as np
import wandb
from IPython import display
from numpy.typing import NDArray

if TYPE_CHECKING:
    from torch import Tensor

TensorOrNDArray = Union["Tensor", NDArray]


def encode_video(
    frames: TensorOrNDArray,
    normalize: bool = False,
    fps: int = 10,
    video_format: str = "webm",
) -> str:
    """
    This is a stripped down version of wandb_lib.log_video. Can pull in more options from there as needed,
    like batching or normalization.
    Keeping them separate so we don't need Torch here.
    Expects an array of shape (t c h w), in uint range (0, 255) or float range (0, 1).
    I think wandb.Video works with a batch of videos natively, too.
    - `format` accepts "gif", "mp4", "webm" or "ogg"
    """
    if normalize:
        raise NotImplementedError

    if isinstance(frames, np.ndarray):
        if np.issubdtype(frames.dtype, np.floating):
            frames = (frames * 255).clip(0, 255).astype(np.uint8)
    else:
        # Assume Torch tensor
        import torch

        frames = frames.detach()
        if torch.is_floating_point(frames):
            frames = (frames * 255).clamp(0, 255).to(torch.uint8)
        frames = frames.cpu()

    return wandb.Video(frames, format=video_format, fps=fps)._path  # type: ignore


def visualize_arraylike_as_video(
    frames: TensorOrNDArray,
    normalize: bool = False,
    fps: int = 10,
    size: tuple[int, int] = (256, 256),
    video_format: str = "webm",
) -> None:
    """Render a video in a jupyter notebook. See `encode_video` for options.
    - `size`: size to display the video
    """
    video_path = encode_video(frames, normalize=normalize, fps=fps, video_format=video_format)
    display.display(
        display.Video(video_path, embed=True, width=size[0], height=size[1], html_attributes="controls autoplay")
    )



from pathlib import Path
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type

import attr
import numpy as np
from scipy import stats

from avalon.datagen.world_creation.configs.export import ExportConfig
from avalon.datagen.world_creation.configs.task import TaskConfig
from avalon.datagen.world_creation.entities.animals import ALL_PREDATOR_CLASSES
from avalon.datagen.world_creation.entities.animals import Predator
from avalon.datagen.world_creation.tasks.eat import add_food_tree_for_simple_task
from avalon.datagen.world_creation.utils import normalized
from avalon.datagen.world_creation.utils import to_2d_point
from avalon.datagen.world_creation.worlds.creation import create_world_from_constraint
from avalon.datagen.world_creation.worlds.difficulty import normal_distrib_range
from avalon.datagen.world_creation.worlds.difficulty import scale_with_difficulty
from avalon.datagen.world_creation.worlds.difficulty import select_categorical_difficulty
from avalon.datagen.world_creation.worlds.export import export_world
from avalon.datagen.world_creation.worlds.types import CompositionalConstraint
from avalon.datagen.world_creation.worlds.world import World
from avalon.datagen.world_creation.worlds.world_locations import WorldLocations


@attr.s(auto_attribs=True, collect_by_mro=True, hash=True)
class AvoidTaskConfig(TaskConfig):
    # how many predators to spawn at difficulty 0.0 and 1.0 respectively
    # note that each predator may be a different type
    predator_count_easy: int = 1
    predator_count_hard: int = 5
    # the number of predators at a difficulty is not exact, but rather, is sampled from a normal distribution with
    # the mean being proportional to the difficulty, and the following standard deviation
    predator_count_std_dev: float = 0.5
    # the max and min dist between the spawn and goal (ie, how far the agents has to walk to complete the task)
    # larger worlds make more space for us to place the obstacles and predators
    min_dist: float = 13.0
    max_dist: float = 52.0
    # positioning for predators works by figuring out how far away they are, along the path (predator_path_fraction),
    # and then how far off the path they should be spawned (predator_distance_off_path)
    # predator_path_fraction is how far along the path the predators will spawn
    # 0.0 means "they will spawn right on top of you" and 1.0 means "they will spawn right on top of the food"
    # it is harder when they spawn near you
    predator_path_fraction_easy: float = 0.9
    predator_path_fraction_hard: float = 0.1
    predator_path_fraction_std_dev: float = 0.1
    # how far predators start to the side of the path at difficulty 0.0 and 1.0 respectively
    predator_distance_off_path_easy: float = 10.0
    predator_distance_off_path_hard: float = 1.0
    predator_distance_off_path_std_dev: float = 1.0
    # how large the "safety radius" around the player will be. This prevents predators from spawning that close
    # if this is too small, you will probably die before you can get your bearings
    safety_radius_easy: float = 11.0
    safety_radius_hard: float = 8.0


def generate_avoid_task(
    rand: np.random.Generator,
    difficulty: float,
    output_path: Path,
    export_config: ExportConfig,
    task_config: AvoidTaskConfig = AvoidTaskConfig(),
) -> None:
    world, locations, difficulty = create_avoid_obstacle(rand, difficulty, export_config, task_config=task_config)
    world, locations = world.end_height_obstacles(locations, is_accessible_from_water=True)
    world = add_food_tree_for_simple_task(world, locations)
    world = world.add_spawn(rand, difficulty, locations.spawn, locations.goal)
    export_world(output_path, rand, world)


def create_avoid_obstacle(
    rand: np.random.Generator,
    difficulty: float,
    export_config: ExportConfig,
    constraint: Optional[CompositionalConstraint] = None,
    task_config: AvoidTaskConfig = AvoidTaskConfig(),
) -> Tuple[World, WorldLocations, float]:
    predator_types, difficulty = select_predator_types(
        task_config.predator_count_easy,
        task_config.predator_count_hard,
        task_config.predator_count_std_dev,
        difficulty,
        rand,
    )

    current_task_max = task_config.max_dist - (task_config.max_dist - task_config.min_dist) * difficulty
    current_task_delta = current_task_max - task_config.min_dist
    desired_goal_dist = task_config.min_dist + current_task_delta * rand.uniform()

    world, locations = create_world_from_constraint(
        stats.norm(desired_goal_dist, 0.5), rand, difficulty, export_config, constraint
    )

    # how far off the path we create the predator. Closer is harder
    predator_distance_mean = scale_with_difficulty(
        difficulty, task_config.predator_distance_off_path_easy, task_config.predator_distance_off_path_hard
    )

    food_dir_2d = normalized(to_2d_point(locations.goal) - to_2d_point(locations.spawn))
    food_dist_2d = np.linalg.norm(to_2d_point(locations.goal) - to_2d_point(locations.spawn))

    # make the player spawn radius a wee bit bigger for this task
    spawn_safety_radius = scale_with_difficulty(
        difficulty, task_config.safety_radius_easy, task_config.safety_radius_hard
    )
    spawn_dist_sq = world.map.get_dist_sq_to(to_2d_point(locations.spawn))
    nearby = spawn_dist_sq < spawn_safety_radius**2
    is_detail_important_new = world.is_detail_important.copy()
    is_detail_important_new[nearby] = True
    world = attr.evolve(world, is_detail_important=is_detail_important_new)

    for predator_type in predator_types:
        # figure out the randomization parameters for difficulty for this predator
        predator_dist = rand.normal(predator_distance_mean, task_config.predator_distance_off_path_std_dev)
        # how far along the line between food and spawn. Closer is harder
        predator_path_location = normal_distrib_range(
            task_config.predator_path_fraction_easy,
            task_config.predator_path_fraction_hard,
            task_config.predator_path_fraction_std_dev,
            rand,
            difficulty,
        )

        # figure out where to place the predator
        predator_offset_dir_2d = np.array([-food_dir_2d[1], food_dir_2d[0]])
        if rand.random() < 0.5:
            predator_offset_dir_2d *= -1

        path_point_2d = to_2d_point(locations.spawn) + (food_dir_2d * food_dist_2d * predator_path_location)
        predator_point_2d = path_point_2d + predator_dist * predator_offset_dir_2d

        # actually add the predator
        world = world.add_random_predator_near_point(rand, predator_type, predator_point_2d, 1, locations.island)

    return world, locations, difficulty


def select_predator_types(
    min_count: int,
    max_count: int,
    count_std_dev: float,
    difficulty: float,
    rand: np.random.Generator,
    exclude: Iterable[Type[Predator]] = tuple(),
    _FORCED: Optional[Iterable[Type[Predator]]] = None,
) -> Tuple[List[Type[Predator]], float]:
    if _FORCED:
        # ensure hardest is last
        _FORCED_LIST = list(sorted(_FORCED, key=ALL_PREDATOR_CLASSES.index))
        return _FORCED_LIST, difficulty

    acceptable_predators = [pred for pred in ALL_PREDATOR_CLASSES if pred not in exclude]

    max_predator_type, difficulty = select_categorical_difficulty(acceptable_predators, difficulty, rand, _FORCED=None)
    predator_count = round(normal_distrib_range(min_count - 0.45, max_count + 0.45, count_std_dev, rand, difficulty))

    acceptable_predators = acceptable_predators[: acceptable_predators.index(max_predator_type) + 1]
    predator_types = [max_predator_type]
    for i in range(predator_count - 1):
        p_type, _ = select_categorical_difficulty(acceptable_predators, difficulty, rand)
        predator_types.append(p_type)
    return predator_types, difficulty



import math
import sys
from itertools import combinations
from typing import Dict
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import Set
from typing import Tuple

import attr
import networkx as nx
import numpy as np
from matplotlib import patches
from matplotlib import pyplot as plt
from networkx import Graph
from networkx.algorithms.approximation import traveling_salesman_problem
from scipy import stats
from scipy.ndimage import convolve
from scipy.spatial import Delaunay
from scipy.spatial import QhullError

from avalon.common.errors import SwitchError
from avalon.common.utils import only
from avalon.contrib.serialization import Serializable
from avalon.datagen.errors import ImpossibleWorldError
from avalon.datagen.godot_base_types import IntRange
from avalon.datagen.godot_base_types import Vector3
from avalon.datagen.world_creation.debug_plots import IS_DEBUG_VIS
from avalon.datagen.world_creation.geometry import BuildingTile
from avalon.datagen.world_creation.geometry import euclidean_distance
from avalon.datagen.world_creation.geometry import get_triangulation_edges
from avalon.datagen.world_creation.indoor.blocks import make_blocks_from_tiles
from avalon.datagen.world_creation.indoor.building import BuildingAestheticsConfig
from avalon.datagen.world_creation.indoor.components import Entrance
from avalon.datagen.world_creation.indoor.components import Hallway
from avalon.datagen.world_creation.indoor.components import Ladder
from avalon.datagen.world_creation.indoor.components import Ramp
from avalon.datagen.world_creation.indoor.components import Room
from avalon.datagen.world_creation.indoor.components import RoomPlacementError
from avalon.datagen.world_creation.indoor.components import RoomTileGraph
from avalon.datagen.world_creation.indoor.components import Story
from avalon.datagen.world_creation.indoor.components import StoryLink
from avalon.datagen.world_creation.indoor.components import Window
from avalon.datagen.world_creation.indoor.components import add_room_tiles_to_grid
from avalon.datagen.world_creation.indoor.components import generate_room
from avalon.datagen.world_creation.indoor.components import set_link_in_grid
from avalon.datagen.world_creation.indoor.constants import DEFAULT_STORY_HEIGHT
from avalon.datagen.world_creation.indoor.constants import MIN_BUILDING_SIZE
from avalon.datagen.world_creation.indoor.constants import MIN_ROOM_SIZE
from avalon.datagen.world_creation.indoor.constants import Azimuth
from avalon.datagen.world_creation.indoor.constants import Orientation
from avalon.datagen.world_creation.indoor.constants import TileIdentity
from avalon.datagen.world_creation.indoor.constants import WallType
from avalon.datagen.world_creation.indoor.tiles import draw_line_in_grid
from avalon.datagen.world_creation.indoor.tiles import find_exterior_wall_footprints
from avalon.datagen.world_creation.indoor.tiles import get_neighbor_tiles
from avalon.datagen.world_creation.indoor.tiles import visualize_tiles
from avalon.datagen.world_creation.indoor.utils import get_evenly_spaced_centroids
from avalon.datagen.world_creation.indoor.utils import inset_borders
from avalon.datagen.world_creation.indoor.utils import rand_integer
from avalon.datagen.world_creation.indoor.wall_footprint import WallFootprint
from avalon.datagen.world_creation.types import BuildingBoolNP


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class FootprintBuilder(Serializable):
    def build(
        self,
        building_width: int,
        building_length: int,
        story_num: int,
        rand: np.random.Generator,
        footprint_below: Optional[BuildingBoolNP],
    ) -> BuildingBoolNP:
        """returns a `np.array` of shape (building_length, building_width) where True represents buildable tiles"""
        raise NotImplementedError

    @staticmethod
    def inset_footprint(
        footprint: BuildingBoolNP, inset: int = 1, min_leftover_size: int = MIN_BUILDING_SIZE
    ) -> BuildingBoolNP:
        reduced_footprint = footprint.copy()
        for i in range(inset):
            reduced_footprint = inset_borders(reduced_footprint, min_leftover_size)
        return reduced_footprint


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class RectangleFootprintBuilder(FootprintBuilder):
    def build(
        self,
        building_width: int,
        building_length: int,
        story_num: int,
        rand: np.random.Generator,
        footprint_below: Optional[BuildingBoolNP],
    ) -> BuildingBoolNP:
        if story_num != 0:
            assert footprint_below is not None, "Must pass footprint below for non-ground stories"
            return FootprintBuilder.inset_footprint(footprint_below)
        return np.ones((building_length, building_width), dtype=np.bool_)


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class TLShapeFootprintBuilder(FootprintBuilder):
    allow_t_shape: bool = True

    def build(
        self,
        building_width: int,
        building_length: int,
        story_num: int,
        rand: np.random.Generator,
        footprint_below: Optional[BuildingBoolNP],
    ) -> BuildingBoolNP:
        if building_width == building_length == MIN_BUILDING_SIZE:
            raise ImpossibleWorldError("Building too small to form a T/L shape footprint")

        if story_num == 1:
            assert footprint_below is not None, "Must pass footprint below if story_num > 0"
            # Optionally inverting block generation direction gives us more variety
            inverted = rand.choice([True, False])
            footprint_below = footprint_below.copy().T if inverted else footprint_below.copy()
            blocks = make_blocks_from_tiles(footprint_below, solid_tile_value=True)
            min_block_size = MIN_BUILDING_SIZE  # avoid tiny sections that can't have rooms built into them
            block = rand.choice([b for b in blocks if b.x.size >= min_block_size and b.z.size >= min_block_size])  # type: ignore[arg-type]
            footprint = np.zeros((building_length, building_width), dtype=np.bool_)
            if inverted:
                tmp = block.x
                block.x = block.z
                block.z = tmp
            footprint[block.z.min_ge : block.z.max_lt, block.x.min_ge : block.x.max_lt] = 1
            return footprint
        elif story_num > 1:
            assert footprint_below is not None, "Must pass footprint below if story_num > 0"
            return FootprintBuilder.inset_footprint(footprint_below)

        rectangles = []
        footprint = np.zeros((building_length, building_width), dtype=np.bool_)

        min_width = max(round(building_width * 0.4), MIN_BUILDING_SIZE)
        max_width = max(round(building_width * 0.8), MIN_BUILDING_SIZE)

        rectangle_width = rand_integer(rand, min_width, max_width)
        if building_length > building_width and self.allow_t_shape:
            x_offset = rand.choice([0, building_width - rectangle_width])
        else:
            x_offset = rand_integer(rand, 0, building_width - rectangle_width)
        rectangles.append((x_offset, 0, rectangle_width, building_length))

        min_length = max(round(building_length * 0.4), MIN_BUILDING_SIZE)
        max_length = max(round(building_length * 0.8), MIN_BUILDING_SIZE)
        rectangle_length = rand_integer(rand, min_length, max_length)
        if building_width > building_length and self.allow_t_shape:
            z_offset = rand.choice([0, building_length - rectangle_length])
        else:
            z_offset = rand_integer(rand, 0, building_length - rectangle_length)
        rectangles.append((0, z_offset, building_width, rectangle_length))

        for x_offset, z_offset, building_width, building_length in rectangles:
            footprint[z_offset : z_offset + building_length, x_offset : x_offset + building_width] = 1
        return footprint


def _smooth(array: np.ndarray) -> None:
    raveled = array.ravel()
    cuts = np.where(np.diff(raveled) != 0)[0] + 1
    chunks = np.split(raveled, cuts)
    for chunk in chunks:
        if (chunk == 1).all() and chunk.size <= 2:
            chunk[True] = 0


def smooth_footprint(footprint: np.ndarray) -> None:
    """
    Removes sections of the footprint that are effectively zero-width (left wall
    is adjacent to right wall / top wall is adjacent to bottom wall).
    """
    _smooth(footprint)
    transposed_footprint = footprint.T.copy()
    _smooth(transposed_footprint)
    footprint[transposed_footprint.T == 0] = 0


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class IrregularFootprintBuilder(FootprintBuilder):
    def build(
        self,
        building_width: int,
        building_length: int,
        story_num: int,
        rand: np.random.Generator,
        footprint_below: Optional[BuildingBoolNP],
    ) -> BuildingBoolNP:
        if building_width == building_length == MIN_BUILDING_SIZE:
            raise ImpossibleWorldError("Building too small to form an irregular shape footprint")

        if story_num > 0:
            assert footprint_below is not None, "Must pass footprint below if story_num > 0"
            footprint = FootprintBuilder.inset_footprint(footprint_below)
            smooth_footprint(footprint)
            return footprint

        footprint = np.zeros((building_length, building_width), dtype=np.bool_)
        rectangles: List[Tuple[int, int, int, int]] = []
        for i in range(5):
            min_width = max(round(building_width * 0.5), MIN_BUILDING_SIZE)
            max_width = max(round(building_width * 0.9), MIN_BUILDING_SIZE)
            min_length = max(round(building_length * 0.5), MIN_BUILDING_SIZE)
            max_length = max(round(building_length * 0.9), MIN_BUILDING_SIZE)

            rectangle_width = rand_integer(rand, min_width, max_width)
            rectangle_length = rand_integer(rand, min_length, max_length)

            if i == 0:
                x_offset = rand_integer(rand, 0, building_width - rectangle_width)
                z_offset = rand_integer(rand, 0, building_length - rectangle_length)
            else:
                another_rectangle = rand.choice(rectangles)
                possible_x_offsets = [another_rectangle[0]]
                another_rectangle_width = another_rectangle[2]
                if rectangle_width < another_rectangle_width:
                    possible_x_offsets.append(another_rectangle_width - rectangle_width)

                possible_z_offsets = [another_rectangle[1]]
                another_rectangle_length = another_rectangle[3]
                if rectangle_length < another_rectangle_length:
                    possible_z_offsets.append(another_rectangle_length - rectangle_length)

                aligned_offset = rand.choice(["x", "z"])
                if aligned_offset == "x":
                    x_offset = rand.choice(possible_x_offsets)
                    z_offset = rand_integer(rand, 0, building_length - rectangle_length)
                else:
                    x_offset = rand_integer(rand, 0, building_width - rectangle_width)
                    z_offset = rand.choice(possible_z_offsets)
            rectangles.append((x_offset, z_offset, rectangle_width, rectangle_length))

        for x_offset, z_offset, building_width, building_length in rectangles:
            footprint[z_offset : z_offset + building_length, x_offset : x_offset + building_width] = 1
        return footprint


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class RoomBuilder(Serializable):
    def build(self, story_footprint: np.ndarray, rand: np.random.Generator) -> List[Room]:
        raise NotImplementedError


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class DefaultRoomBuilder(RoomBuilder):
    room_count: int
    room_options: List[Vector3]

    def build(self, story_footprint: np.ndarray, rand: np.random.Generator) -> List[Room]:
        world_length, world_width = story_footprint.shape

        rooms = []
        room_floor_space = 0
        for i in range(self.room_count):
            width, height, length = rand.choice(self.room_options)  # type: ignore[arg-type]
            room = generate_room(i + 1, width, height, length)
            assert room.width < world_width and room.length < world_length, "Room too large to fit world"
            rooms.append(room)
            room_floor_space += width * length
        assert room_floor_space < world_width * world_length, "Room floor space exceeds world size"
        return self._place_rooms(story_footprint, rooms)

    def _place_rooms(self, story_footprint: BuildingBoolNP, rooms: List[Room]) -> List[Room]:
        """
        At its core, this is a rectangle packing problem, which is NP-hard. There are many published algorithms out there
        that we could use if we want to, but the one below is a quick and dirty homemade one while we iterate this.
        """
        grid = np.invert(story_footprint).astype(np.int64)
        placed_rooms = []
        sorted_rooms = sorted(rooms, key=lambda r: r.floor_space, reverse=True)
        blocks = make_blocks_from_tiles(grid)
        margin = 1
        for i, room in enumerate(sorted_rooms):
            for block in blocks:
                block_length = block.z.max_lt - block.z.min_ge
                block_width = block.x.max_lt - block.x.min_ge
                if block_width < room.width + (margin * 2) or block_length < room.length + (margin * 2):
                    continue
                center = block.x.min_ge + block_width // 2, block.z.min_ge + block_length // 2
                room = room.with_position(
                    BuildingTile(x=center[0] - room.width // 2, z=center[1] - room.length // 2)
                ).with_id(i)
                add_room_tiles_to_grid(room, room.tiles, grid)
                placed_rooms.append(room)
                blocks = make_blocks_from_tiles(grid)
                break
            else:
                # The options here are: a) try a more compact placement algorithm; b) skip room; for now we just raise
                visualize_tiles(grid)
                raise RoomPlacementError(
                    f"can't fit room of size {room.width, room.length} with {margin=} in any block"
                )
        return placed_rooms


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class HouseLikeRoomBuilder(RoomBuilder):
    min_room_size: int = 3  # Note: this param will not be obeyed if the building is too small to fit
    max_rooms: int = sys.maxsize
    partition_thickness: int = 1
    room_height: float = DEFAULT_STORY_HEIGHT

    def build(self, story_footprint: np.ndarray, rand: np.random.Generator) -> List[Room]:
        world_length, world_width = story_footprint.shape
        random_state = np.random.RandomState(seed=rand.integers(0, 2**32 - 1))

        # Separate footprint into rectangles as "initial rooms"
        padded_footprint = story_footprint.copy()
        for i in range(self.partition_thickness):
            # We want rooms to be at least of width 2 to be able to place story links that we can walk around
            padded_footprint = inset_borders(padded_footprint, min_leftover_size=MIN_ROOM_SIZE)
        footprint_blocks = make_blocks_from_tiles(padded_footprint, solid_tile_value=True)
        initial_rooms = [
            Room(-1, BuildingTile(block.x.min_ge, block.z.min_ge), block.x.size, block.z.size, self.room_height)
            for block in footprint_blocks
        ]

        if len(initial_rooms) == 1 and initial_rooms[0].width == 1 or initial_rooms[0].length == 1:
            # Don't make rooms that can't have walls built between them anyway
            return [room.with_id(i) for i, room in enumerate(initial_rooms)]

        initial_orientation = rand.choice(list(Orientation))  # type: ignore[arg-type]

        # Then partition each of the rooms
        rooms = []
        if len(initial_rooms) > self.max_rooms:
            raise ImpossibleWorldError(
                f"max_rooms is {self.max_rooms}, but initial footprint split yields {len(initial_rooms)}"
            )
        elif len(initial_rooms) == self.max_rooms:
            rooms = initial_rooms
        else:
            leftover_credits = self.max_rooms - len(initial_rooms)  # how many more rooms can you create
            for room in initial_rooms:
                new_rooms, leftover_credits = self._partition(
                    room,
                    initial_orientation,
                    world_width,
                    world_length,
                    self.partition_thickness,
                    leftover_credits,
                    random_state,
                )
                rooms.extend(new_rooms)
        return [room.with_id(i) for i, room in enumerate(rooms)]

    def _partition(
        self,
        room: Room,
        orientation: Orientation,
        world_width: int,
        world_length: int,
        partition_thickness: int,
        leftover_room_credits: int,
        random_state: np.random.RandomState,
    ) -> Tuple[List[Room], int]:
        if leftover_room_credits == 0:
            return [room], leftover_room_credits

        if orientation == Orientation.HORIZONTAL:
            edge_size = room.width
        else:
            edge_size = room.length
        if edge_size < 2 * self.min_room_size + partition_thickness:
            return [room], leftover_room_credits
        partition_location_distribution = stats.uniform(
            self.min_room_size, edge_size - 2 * self.min_room_size - partition_thickness
        )
        partition_location = round(partition_location_distribution.rvs(random_state=random_state))
        second_room_offset = partition_location + partition_thickness
        if orientation == Orientation.HORIZONTAL:
            partition_a = Room(
                -1,
                BuildingTile(room.position.x, room.position.z),
                width=partition_location,
                length=room.length,
                outer_height=room.outer_height,
            )
            partition_b = Room(
                -1,
                BuildingTile(room.position.x + second_room_offset, room.position.z),
                width=room.width - second_room_offset,
                length=room.length,
                outer_height=room.outer_height,
            )

        else:
            partition_a = Room(
                -1,
                BuildingTile(room.position.x, room.position.z),
                width=room.width,
                length=partition_location,
                outer_height=room.outer_height,
            )
            partition_b = Room(
                -1,
                BuildingTile(room.position.x, room.position.z + second_room_offset),
                width=room.width,
                length=room.length - second_room_offset,
                outer_height=room.outer_height,
            )
        next_partition_dimension_size = world_width if orientation == Orientation.HORIZONTAL else world_length
        probability_keep_partitioning = edge_size / next_partition_dimension_size
        keep_partitioning_probabilities = [probability_keep_partitioning, 1 - probability_keep_partitioning]
        keep_partitioning_a = random_state.choice([True, False], p=keep_partitioning_probabilities)
        keep_partitioning_b = random_state.choice([True, False], p=keep_partitioning_probabilities)
        orientation = orientation.other()
        leftover_room_credits -= 1
        partitioned_a, leftover_room_credits = (
            self._partition(
                partition_a,
                orientation,
                world_width,
                world_length,
                partition_thickness,
                leftover_room_credits,
                random_state,
            )
            if keep_partitioning_a
            else ([partition_a], leftover_room_credits)
        )
        partitioned_b, leftover_room_credits = (
            self._partition(
                partition_b,
                orientation,
                world_width,
                world_length,
                partition_thickness,
                leftover_room_credits,
                random_state,
            )
            if keep_partitioning_b
            else ([partition_b], leftover_room_credits)
        )
        return [*partitioned_a, *partitioned_b], leftover_room_credits


def get_room_overlap(room_a: Room, room_b: Room) -> Tuple[Optional[IntRange], Optional[IntRange]]:
    return room_a.x_range.overlap(room_b.x_range), room_a.z_range.overlap(room_b.z_range)


def are_rooms_neighboring(room_a: Room, room_b: Room, max_gap: int = 1) -> bool:
    x_overlap, z_overlap = get_room_overlap(room_a, room_b)
    if x_overlap is not None:
        topmost_bottom = min(room_a.position.z + room_a.length, room_b.position.z + room_b.length) - 1
        bottommost_top = max(room_a.position.z, room_b.position.z)
        # The gap is one less than index difference (adjacent room idx delta is 1, but the "gap" is 0)
        gap = bottommost_top - topmost_bottom - 1
        return gap <= max_gap
    elif z_overlap is not None:
        leftmost_right = min(room_a.position.x + room_a.width, room_b.position.x + room_b.width) - 1
        rightmost_left = max(room_a.position.x, room_b.position.x)
        gap = rightmost_left - leftmost_right - 1
        return gap <= max_gap
    else:
        return False


def get_hallway_attachment_points(room_a: Room, room_b: Room) -> Tuple[BuildingTile, BuildingTile]:
    overlap_x, overlap_z = get_room_overlap(room_a, room_b)
    rightmost_room = room_a if room_a.center.x > room_b.center.x else room_b
    leftmost_room = room_a if rightmost_room == room_b else room_b
    bottommost_room = room_a if room_a.center.z > room_b.center.z else room_b
    topmost_room = room_a if bottommost_room == room_b else room_b
    if overlap_x:
        x_left = x_right = math.floor(overlap_x.midpoint)
        z_top = topmost_room.z_range.max_lt - 1
        z_bottom = bottommost_room.position.z
    elif overlap_z:
        z_top = z_bottom = math.floor(overlap_z.midpoint)
        x_left = leftmost_room.x_range.max_lt - 1
        x_right = rightmost_room.position.x
    else:
        x_left = min(room_a.x_range.max_lt, room_b.x_range.max_lt) - 1
        x_right = max(room_a.x_range.min_ge, room_b.x_range.min_ge)
        z_top = min(room_a.z_range.max_lt, room_b.z_range.max_lt) - 1
        z_bottom = max(room_a.z_range.min_ge, room_b.z_range.min_ge)

    attachment_a_z, attachment_b_z = (z_top, z_bottom) if topmost_room == room_a else (z_bottom, z_top)
    attachment_a_x, attachment_b_x = (x_left, x_right) if leftmost_room == room_a else (x_right, x_left)
    room_a_attachment_point = BuildingTile(attachment_a_x, attachment_a_z)
    room_b_attachment_point = BuildingTile(attachment_b_x, attachment_b_z)
    return room_a_attachment_point, room_b_attachment_point


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class HallwayBuilder(Serializable):
    def build(self, rooms: List[Room], rand: np.random.Generator) -> List[Hallway]:
        raise NotImplementedError


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class NoHallwayBuilder(HallwayBuilder):
    """For testing only - you want connections between rooms!"""

    def build(self, rooms: List[Room], rand: np.random.Generator) -> List[Hallway]:
        return []


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class DefaultHallwayBuilder(HallwayBuilder):
    proportion_additional_edges: float = 0.18

    def _get_room_centroids(self, rooms: List[Room]) -> Dict[int, Tuple[int, int]]:
        return {i: (r.center.x, r.center.z) for i, r in enumerate(rooms)}

    def _make_complete_room_graph(self, rooms: List[Room]) -> Graph:
        graph = nx.complete_graph(len(rooms))
        centroids_by_room = self._get_room_centroids(rooms)
        nx.set_node_attributes(graph, centroids_by_room, "pos")
        distance_by_rooms = {}
        for i, room_a in enumerate(rooms):
            for j, room_b in enumerate(rooms):
                distance_by_rooms[(i, j)] = euclidean_distance(room_a.center, room_b.center)
        nx.set_edge_attributes(graph, distance_by_rooms, "distance")
        return graph

    def _make_final_room_graph(self, rooms: List[Room], rand: np.random.Generator) -> Graph:
        graph = self._make_complete_room_graph(rooms)
        mst = nx.minimum_spanning_tree(graph, weight="distance")
        centroids_by_room = self._get_room_centroids(rooms)
        room_centroids = np.array([list(p) for p in centroids_by_room.values()])
        if len(rooms) > 2:
            try:
                room_triangulation = Delaunay(room_centroids)
            except QhullError:
                # This means the rooms are in a vertical/horizontal line and won't have meaningful extra connections
                return mst
            edges = get_triangulation_edges(room_triangulation)
            unjoined_edges = list(edges - set(mst.edges))
            additional_edge_count = round(self.proportion_additional_edges * len(unjoined_edges))
            additional_edges = rand.choice(unjoined_edges, additional_edge_count, replace=False)
            for edge_start, edge_end in additional_edges:
                mst.add_edge(edge_start, edge_end)
        return mst

    def build(self, rooms: List[Room], rand: np.random.Generator) -> List[Hallway]:
        if len(rooms) == 1:
            return []

        connection_graph = self._make_final_room_graph(rooms, rand)
        tile_graph = self._make_tile_graph(rooms)

        hallways = []
        for from_idx, to_idx in connection_graph.edges:
            from_room, to_room = rooms[from_idx], rooms[to_idx]
            hallway = self._build_hallway(tile_graph, from_room, to_room)
            hallways.append(hallway)
        return hallways

    def _build_hallway(self, tile_graph: RoomTileGraph, from_room: Room, to_room: Room) -> Hallway:
        from_room_point, to_room_point = get_hallway_attachment_points(from_room, to_room)
        path = tile_graph.find_shortest_path(from_room_point, to_room_point)

        optimized_path = [path[0]]
        for previous_point, this_point, next_point in zip(path[:-2], path[1:-1], path[2:]):
            previous_direction = Orientation.HORIZONTAL if this_point[0] == previous_point[0] else Orientation.VERTICAL
            next_direction = Orientation.HORIZONTAL if this_point[0] == next_point[0] else Orientation.VERTICAL
            if previous_direction != next_direction:
                optimized_path.append(this_point)
        optimized_path.append(path[-1])
        points = tuple(BuildingTile(x=pt[1], z=pt[0]) for pt in optimized_path)
        return Hallway(points, from_room.id, to_room.id, width=1)

    def _make_tile_graph(self, rooms: List[Room]) -> RoomTileGraph:
        return RoomTileGraph(rooms)


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class HamiltonianHallwayBuilder(DefaultHallwayBuilder):
    """
    Tries to create hallways such that they form a Hamiltonian path between all the rooms.

    Not all graphs have a Hamiltonian path and the implementation uses a heuristic-based TSP solutions, so the resulting
    path may not necessarily be Hamiltonian - but it's the closest we can trivially get without a brute-force solution.

    The builder checks for self-loops in the path (which would mean hallways traverse through other rooms), and raises
    an ImpossibleWorldError if all methods fail to create a path without self-loops.

    NOTE: This hallway builder is only compatible with HouseLikeRoomBuilder-built rooms.
    """

    def _make_final_room_graph(self, rooms: List[Room], rand: np.random.Generator) -> Graph:
        graph = self._make_complete_room_graph(rooms)
        for from_node, to_node in combinations(graph.nodes, 2):
            from_room = rooms[from_node]
            to_room = rooms[to_node]
            if not are_rooms_neighboring(from_room, to_room, max_gap=1):
                graph.remove_edge(from_node, to_node)

        if len(rooms) == 1:
            return graph
        path = traveling_salesman_problem(graph, cycle=False)
        is_path_hamiltonian = len(path) == len(set(path)) and len(path) == len(rooms)
        if not is_path_hamiltonian:
            raise ImpossibleWorldError("TSP solution yielded non-Hamiltonian paths")
        final_graph = nx.create_empty_copy(graph)  # keeps nodes, removes edges
        for from_node, to_node in zip(path[:-1], path[1:]):
            final_graph.add_edge(from_node, to_node)
        return final_graph


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class StoryLinker(Serializable):
    def link_stories(self, stories: List[Story], rand: np.random.Generator) -> List[StoryLink]:
        raise NotImplementedError


class CantFitRampError(ImpossibleWorldError):
    pass


class CantFitLadderError(ImpossibleWorldError):
    pass


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class DefaultStoryLinker(StoryLinker):
    raise_on_failure: bool = True
    allow_ramps: bool = True
    allow_ladders: bool = True

    def link_stories(self, stories: List[Story], rand: np.random.Generator) -> List[StoryLink]:
        links = []
        for bottom_story, top_story in zip(stories[:-1], stories[1:]):
            link = None
            if self.allow_ramps:
                try:
                    link = self.make_ramp(bottom_story, top_story, rand)
                except CantFitRampError:
                    pass

            if link is None and self.allow_ladders:
                link = self.make_ladder(bottom_story, top_story, rand)

            if link is None:
                if self.raise_on_failure:
                    raise ImpossibleWorldError(f"Could not link stories {bottom_story.num} and {top_story.num}")
                else:
                    continue
            links.append(link)
        return links

    def try_fit_ramp(
        self,
        bottom_story: Story,
        top_story: Story,
        bottom_room: Room,
        top_room: Room,
        bottom_story_tile_z: int,
        bottom_story_tile_x: int,
        bottom_story_tile_empty: bool,
        ramp_block: np.ndarray,
        ramp_mask: BuildingBoolNP,
        top_landing_mask: BuildingBoolNP,
        reverse: bool,
    ) -> Optional[Ramp]:
        ramp_block_length, ramp_block_width = ramp_block.shape

        if not bottom_story_tile_empty:
            return None

        z, x = bottom_story_tile_z, bottom_story_tile_x
        if reverse:
            ramp_within_bounds = z - ramp_block_length + 1 >= 0 and x - ramp_block_width + 1 >= 0
        else:
            ramp_within_bounds = (
                z + ramp_block_length - 1 <= ramp_mask.shape[0] and x + ramp_block_width - 1 <= ramp_mask.shape[1]
            )
        if not ramp_within_bounds:
            return None

        if reverse:
            ramp_space_free = (
                ramp_mask[z - ramp_block_length + 1 : z + 1, x - ramp_block_width + 1 : x + 1] == 1
            ).all()
        else:
            ramp_space_free = (ramp_mask[z : z + ramp_block_length, x : x + ramp_block_width] == 1).all()
        if not ramp_space_free:
            return None

        # TODO: check top landing properly when doing 1d9f913a-3870-41b0-839e-8b29d0b89983
        if reverse:
            top_landing_coords = z - ramp_block_length + 1, x - ramp_block_width + 1
        else:
            top_landing_coords = z + ramp_block_length - 1, x + ramp_block_width - 1
        top_landing_in_bounds = (
            0 <= top_landing_coords[0] < top_landing_mask.shape[0]
            and 0 <= top_landing_coords[1] < top_landing_mask.shape[1]
        )
        if not top_landing_in_bounds:
            return None

        top_landing_space_free = top_landing_mask[top_landing_coords] == 1
        if not top_landing_space_free:
            return None

        return Ramp(
            bottom_story.num,
            bottom_room.id,
            BuildingTile(x, z),
            top_story.num,
            top_room.id,
            BuildingTile(top_landing_coords[1], top_landing_coords[0]),
            width=1,
        )

    def make_ladder(self, bottom_story: Story, top_story: Story, rand: np.random.Generator) -> Optional[Ladder]:
        azimuths: List[Azimuth] = list(Azimuth)
        rand.shuffle(azimuths)  # type: ignore

        for azimuth in azimuths:
            bottom_story_free_mask = _get_story_free_mask(bottom_story)
            top_story_free_mask = _get_story_free_mask(top_story)

            # Ladders require two tiles: the principal tile where the ladder is placed + an adjacent tile one step
            # backwards from the ladder to make room for the fat player. The azimuth mask tells us if that tile is free
            backwards_adjacent_tile_kernel = azimuth.convolution_kernel
            forwards_adjacent_tile_kernel = azimuth.opposite.convolution_kernel
            # Top story needs a tile forward from ladder to "land" on
            has_top_landing_tile_mask = convolve(top_story_free_mask, forwards_adjacent_tile_kernel)
            ladder_site_mask = bottom_story_free_mask & top_story_free_mask
            # Both stories need a tile backwards as it'll be cut out
            backwards_adjacent_tile_mask = convolve(ladder_site_mask, backwards_adjacent_tile_kernel)
            ladder_site_mask &= backwards_adjacent_tile_mask & has_top_landing_tile_mask

            viable_tiles = list(zip(*np.where(ladder_site_mask)))
            if not viable_tiles:
                # For narrow buildings, only horizontal or vertical azimuths might work out
                continue

            ladder_tile = rand.choice(viable_tiles)
            ladder_position = BuildingTile(x=ladder_tile[1], z=ladder_tile[0])

            top_position = ladder_position
            if azimuth == Azimuth.NORTH:
                bottom_position = attr.evolve(top_position, z=top_position.z + 1)
            elif azimuth == Azimuth.EAST:
                bottom_position = attr.evolve(top_position, x=top_position.x - 1)
            elif azimuth == Azimuth.SOUTH:
                bottom_position = attr.evolve(top_position, z=top_position.z - 1)
            elif azimuth == Azimuth.WEST:
                bottom_position = attr.evolve(top_position, x=top_position.x + 1)
            else:
                raise SwitchError(f"Unknown azimuth {azimuth}")

            bottom_room = bottom_story.get_room_at_point(bottom_position)
            top_room = top_story.get_room_at_point(top_position)
            assert bottom_room is not None and top_room is not None

            # TODO: 1d9f913a-3870-41b0-839e-8b29d0b89983
            return Ladder(
                bottom_story_id=bottom_story.num,
                bottom_room_id=bottom_room.id,
                bottom_position=bottom_position,
                top_story_id=top_story.num,
                top_room_id=top_room.id,
                top_position=top_position,
                width=1,
                azimuth=azimuth,
            )

        if not self.raise_on_failure:
            return None
        raise CantFitLadderError(
            "Stories don't have enough overlap to place ladder."
            "Either something is very wrong or the building is too small"
        )

    def make_ramp(self, bottom_story: Story, top_story: Story, rand: np.random.Generator) -> Optional[StoryLink]:
        """
        The rough algorithm for connecting two stories with a ramp:
        1) Iterate over all bottom story rooms, starting with the biggest bottom one
        2) Iterate over all top story rooms, starting with nearest to current bottom one
        3) Calculate:
            bottom landing mask (which tiles can we place a landing on the bottom floor?)
            top landing mask
            ramp mask (basically all tiles except for rooms that we are not connecting and all hallways+landings)
        4) Iterate tile-by-tile in free bottom landing space
        5) Iterate over the 4 possible directions
        6) Try placing the ramp w/ landings: if ramp space and landing spaces are all clear, we can place the ramp!
        """
        assert (
            bottom_story.width == top_story.width and bottom_story.length == bottom_story.length
        ), "Different-sized story linking not implemented"

        for bottom_room in sorted(bottom_story.rooms, key=lambda room: room.tiles.size):
            for top_room in sorted(
                top_story.rooms, key=lambda top_room: euclidean_distance(top_room.center, bottom_room.center)
            ):
                ramp = self._try_link_rooms_by_ramp(bottom_story, top_story, bottom_room, top_room)
                if ramp is not None:
                    return ramp

        if not self.raise_on_failure:
            return None
        else:
            raise CantFitRampError("Building too small to place ramp")

    def _try_link_rooms_by_ramp(
        self, bottom_story: Story, top_story: Story, bottom_room: Room, top_room: Room
    ) -> Optional[Ramp]:
        ramp_width = 1
        landing_length = 1
        max_ramp_angle = math.radians(45)
        vertical_distance = bottom_story.outer_height + top_story.floor_negative_depth
        rounding_error_threshold_decimals = 2
        ramp_floor_length = math.ceil(
            round(vertical_distance / math.tan(max_ramp_angle), rounding_error_threshold_decimals)
        )
        required_space = (ramp_floor_length + 2 * landing_length, ramp_width)
        ramp_block = np.empty(required_space)

        bottom_room_hallways = [h for h in bottom_story.hallways if h.from_room_id == bottom_room.id]
        bottom_landing_mask = _get_room_free_mask(
            bottom_story.width,
            bottom_story.length,
            bottom_room,
            bottom_room_hallways,
            bottom_story.story_links,
        )
        top_room_hallways = [h for h in top_story.hallways if h.from_room_id == top_room.id]
        top_landing_mask = _get_room_free_mask(
            top_story.width, top_story.length, top_room, top_room_hallways, top_story.story_links
        )
        if not (bottom_landing_mask == 1).any():
            return None
        if not (top_landing_mask == 1).any():
            return None

        # free space = ones
        other_bottom_rooms = [room for room in bottom_story.rooms if room != bottom_room]
        other_top_rooms = [room for room in top_story.rooms if room != top_room]
        ramp_mask = _get_ramp_free_mask(
            bottom_story.width,
            bottom_story.length,
            exclude_rooms=[*other_bottom_rooms, *other_top_rooms],
            exclude_hallways=[*bottom_story.hallways, *top_story.hallways],
            exclude_story_links=[*bottom_story.story_links, *top_story.story_links],
        )

        for ramp_block in [ramp_block, ramp_block.T]:
            ramp_block_length, ramp_block_width = ramp_block.shape
            max_free_length = (ramp_mask == 1).sum(axis=0).max()
            max_free_width = (ramp_mask == 1).sum(axis=1).max()
            if ramp_block_length > max_free_length or ramp_block_width > max_free_width:
                continue
            for reverse in [False, True]:
                for (z, x), empty in np.ndenumerate(bottom_landing_mask == 1):
                    ramp = self.try_fit_ramp(
                        bottom_story,
                        top_story,
                        bottom_room,
                        top_room,
                        bottom_story_tile_z=z,
                        bottom_story_tile_x=x,
                        bottom_story_tile_empty=empty,
                        ramp_block=ramp_block,
                        ramp_mask=ramp_mask,
                        top_landing_mask=top_landing_mask,
                        reverse=reverse,
                    )
                    if ramp:
                        return ramp
        return None


def _get_room_free_mask(
    story_width: int, story_length: int, room: Room, hallways: List[Hallway], story_links: List[StoryLink]
) -> BuildingBoolNP:
    """Returns a mask where ones represent the free space in a room (free of ramps, landings, etc.)"""
    grid = np.zeros((story_length, story_width), dtype=np.float32)
    add_room_tiles_to_grid(room, room.tiles, grid)
    grid = grid.astype(bool)
    for hallway in hallways:
        grid[hallway.points[0].z, hallway.points[0].x] = 0
        grid[hallway.points[-1].z, hallway.points[-1].x] = 0
    for story_link in story_links:
        draw_line_in_grid((story_link.bottom_position, story_link.top_position), grid, 0, include_ends=True)
    return grid


def _get_story_free_mask(story: Story) -> BuildingBoolNP:
    story_free_mask = np.zeros((story.length, story.width), dtype=np.bool_)
    for room in story.rooms:
        room_free_mask = _get_room_free_mask(story.width, story.length, room, story.hallways, story.story_links)
        story_free_mask |= room_free_mask
    return story_free_mask.astype(np.bool_)


def _get_ramp_free_mask(
    story_width: int,
    story_length: int,
    exclude_rooms: List[Room],
    exclude_hallways: List[Hallway],
    exclude_story_links: List[StoryLink],
) -> BuildingBoolNP:
    """
    Returns a mask where ones represent the free space for building a ramp across multiple floors
    """
    grid = np.zeros((story_length, story_width), dtype=np.int_)
    for room in exclude_rooms:
        add_room_tiles_to_grid(room, room.tiles, grid)
    for hallway in exclude_hallways:
        draw_line_in_grid(hallway.points, grid, TileIdentity.HALLWAY.value, include_ends=True)
    for story_link in exclude_story_links:
        draw_line_in_grid(
            (story_link.bottom_position, story_link.top_position), grid, TileIdentity.LINK.value, include_ends=True
        )
    return np.invert(grid.astype(bool))


class ObstacleSite(NamedTuple):
    story_num: int
    room_id: int
    site_index: int  # not called index to avoid collision with NamedTuple.index
    length: int
    vertical: bool


def get_free_obstacle_sites(stories: List[Story]) -> List[ObstacleSite]:
    options = []
    for story in stories:
        for room in story.rooms:
            # TODO (1d9f913a-3870-41b0-839e-8b29d0b89983): support sites with width > 1
            # Gets all free verticals/horizontals, possibly in front of hallway joints
            global_room_free_mask = _get_room_free_mask(
                story.width, story.length, room, hallways=[], story_links=story.story_links
            )
            local_room_free_mask = global_room_free_mask[
                room.z_range.min_ge : room.z_range.max_lt, room.x_range.min_ge : room.x_range.max_lt
            ]
            free_vertical_idxs = only(np.where((local_room_free_mask == 1).all(axis=0)))
            options.extend([ObstacleSite(story.num, room.id, idx, room.length, True) for idx in free_vertical_idxs])
            free_horizontal_idxs = only(np.where((local_room_free_mask == 1).all(axis=1)))
            options.extend([ObstacleSite(story.num, room.id, idx, room.width, False) for idx in free_horizontal_idxs])
    return options


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class EntranceBuilder:
    def build(self, stories: List[Story], rand: np.random.Generator) -> List[Story]:
        raise NotImplementedError


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class DefaultEntranceBuilder(EntranceBuilder):
    top: bool
    bottom: bool

    def build(self, stories: List[Story], rand: np.random.Generator) -> List[Story]:
        if self.top:
            # TODO: 8451efda-5f27-4ea3-a1d3-cde7f84f652c
            # self.add_entrance_ramp(stories[-1], rand)
            raise NotImplementedError
        if self.bottom:
            first_story = stories[0]
            self.add_entrance(first_story, rand)
        return stories

    def add_entrance(
        self,
        story: Story,
        rand: np.random.Generator,
        permitted_room_ids: Optional[List[int]] = None,
        permitted_azimuths: Optional[List[Azimuth]] = None,
    ) -> Entrance:
        entrance = self.make_entrance(story, rand, permitted_room_ids, permitted_azimuths)
        self.remove_windows_in_way_of_entrance(story, entrance)
        # TODO: think about how to do entrance/hallway relationship better
        story.entrances.append(entrance)
        story.hallways.append(entrance.hallway)
        return entrance

    def remove_windows_in_way_of_entrance(self, story: Story, entrance: Entrance) -> None:
        entrance_points_in_outline = set(entrance.get_points_in_story_outline(story))
        adjacent_points: Set[Tuple[int, int]] = set()
        for point in entrance_points_in_outline:
            adjacent_points.update(get_neighbor_tiles(story.get_outline_mask(), point, only_if_equals=True))
        entrance_points_in_outline.update(adjacent_points)

        for entry_z, entry_x in sorted(list(entrance_points_in_outline)):
            window_idxs_to_remove = []
            for i, window in enumerate(story.windows):
                window_x, _, window_z = window.position
                window_x_range = IntRange(
                    window.position[0] - window.size[0] / 2, window.position[0] + window.size[0] / 2
                )
                window_z_range = IntRange(
                    window.position[2] - window.size[2] / 2, window.position[2] + window.size[2] / 2
                )
                entry_tile_x_range = IntRange(entry_x, entry_x + 1)
                entry_tile_z_range = IntRange(entry_z, entry_z + 1)
                if entry_tile_x_range.overlap(window_x_range) and entry_tile_z_range.overlap(window_z_range):
                    window_idxs_to_remove.append(i)
            for idx in sorted(window_idxs_to_remove, reverse=True):
                del story.windows[idx]

    def make_entrance(
        self,
        story: Story,
        rand: np.random.Generator,
        permitted_room_ids: Optional[List[int]] = None,
        permitted_azimuths: Optional[List[Azimuth]] = None,
    ) -> Entrance:
        permitted_azimuths = list(Azimuth) if permitted_azimuths is None else permitted_azimuths
        tiles: np.ndarray = story.generate_tiles(include_hallway_landings=True)
        room_id_tiles = story.generate_room_id_tiles()
        outline_mask = story.get_outline_mask()
        is_tile_free = tiles == TileIdentity.ROOM.value

        # Find all possible story outline tiles that are not blocked
        # To do this, we create a fitting mask, where all free tiles are 1s (is_tile_free) and where the outline
        # is an arbitrary number other than 1.
        arbitrary_multiplier = 9
        fitting_mask = outline_mask.astype(np.int_) * arbitrary_multiplier + is_tile_free.astype(int)
        viable_connections = []
        for wall_type in WallType:
            azimuth = wall_type.azimuth
            if azimuth not in permitted_azimuths:
                continue

            # To check if a tile is suitable, we check that it is free (is_tile_free) AND that it  has an outline wall
            # of the desired azimuth as a neighbor (through the convolution with the wall kernel).
            # The expected mask value is 3 * arbitrary multiplier, since the wall kernels are all 1s on the wall side.
            kernel = wall_type.convolution_kernel
            assert kernel.shape[0] == kernel.shape[1]
            kernel_size = kernel.shape[0]
            is_suitable_entrance_site = is_tile_free.astype(int) & (
                convolve(fitting_mask, kernel, mode="constant") == kernel_size * arbitrary_multiplier
            )
            suitable_tiles = list(zip(*np.where(is_suitable_entrance_site == True)))
            for tile in suitable_tiles:
                interior_point = BuildingTile(z=tile[0], x=tile[1])
                room_id = room_id_tiles[interior_point.z, interior_point.x]
                if permitted_room_ids is not None and room_id not in permitted_room_ids:
                    continue

                if wall_type == WallType.NORTH:
                    exterior_point = BuildingTile(z=interior_point.z - 2, x=interior_point.x)
                elif wall_type == WallType.EAST:
                    exterior_point = BuildingTile(z=interior_point.z, x=interior_point.x + 2)
                elif wall_type == WallType.SOUTH:
                    exterior_point = BuildingTile(z=interior_point.z + 2, x=interior_point.x)
                elif wall_type == WallType.WEST:
                    exterior_point = BuildingTile(z=interior_point.z, x=interior_point.x - 2)
                else:
                    raise SwitchError("Unexpected wall type")

                viable_connections.append((room_id, azimuth, (exterior_point, interior_point)))

        # TODO: 1d9f913a-3870-41b0-839e-8b29d0b89983
        if len(viable_connections) == 0:
            raise ImpossibleWorldError(f"Can't build entryway to any room ({permitted_room_ids=})")
        connection_idx = rand.choice(list(range(len(viable_connections))))
        connected_room_id, azimuth, hallway_points = viable_connections[connection_idx]

        return Entrance(story.num, connected_room_id, azimuth, hallway_points, width=1)


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class WindowBuilder:
    def build(
        self, stories: List[Story], rand: np.random.Generator, aesthetics: BuildingAestheticsConfig
    ) -> List[Story]:
        stories_with_windows = []
        for story in stories:
            windows = []

            no_build_mask = np.zeros((story.length, story.width))
            for hallway in story.hallways:
                draw_line_in_grid(hallway.points, no_build_mask, TileIdentity.HALLWAY.value, include_ends=False)
            for link in story.story_links:
                set_link_in_grid(
                    link,
                    no_build_mask,
                    set_bottom_landing=link.bottom_story_id == story.num,
                    set_top_landing=link.top_story_id == story.num,
                )
            no_build_mask = no_build_mask.astype(np.bool_)

            wall_footprints = find_exterior_wall_footprints(story.footprint, no_build_mask=no_build_mask)
            for wall_footprint in wall_footprints:
                min_gap = aesthetics.window_min_gap
                window_width = aesthetics.window_width
                window_height = aesthetics.window_height
                window_centroid_y = story.outer_height * aesthetics.window_y_proportion_of_height
                centroids = get_evenly_spaced_centroids(
                    wall_footprint, window_width, min_gap, window_centroid_y, aesthetics.max_windows_per_wall
                )
                windows_for_wall = []
                for centroid in centroids:
                    if wall_footprint.is_vertical:
                        size = np.array([wall_footprint.wall_thickness, window_height, window_width])
                    else:
                        size = np.array([window_width, window_height, wall_footprint.wall_thickness])
                    window = Window(centroid, size)
                    windows_for_wall.append(window)
                windows.extend(windows_for_wall)

            if IS_DEBUG_VIS:
                self.visualise(story, wall_footprints, windows)

            stories_with_windows.append(attr.evolve(story, windows=windows))
        return stories_with_windows

    def visualise(self, story: Story, wall_footprints: List[WallFootprint], windows: List[Window]) -> None:
        fig, ax = plt.subplots()
        plt.gca().set_aspect("equal")
        plt.gca().xaxis.set_major_locator(plt.MultipleLocator(1))
        plt.setp(ax.get_xticklabels(), rotation=90)
        plt.gca().yaxis.set_major_locator(plt.MultipleLocator(1))
        ax.invert_yaxis()
        ax.set_xlim(-1, story.width + 1)
        ax.set_ylim(story.length + 1, -1)
        ax.grid()
        ax.set_title(f"Story {story.num}")

        for footprint in wall_footprints:
            rect = patches.Rectangle(
                footprint.top_left,
                footprint.footprint_width,
                footprint.footprint_length,
                0,
                linewidth=1,
                edgecolor="blue",
                facecolor="none",
            )
            ax.add_patch(rect)
        for window in windows:
            top_left = window.position[0] - window.size[0] / 2, window.position[2] - window.size[2] / 2
            rect = patches.Rectangle(
                top_left, window.size[0], window.size[2], linewidth=2, edgecolor="red", facecolor="none"
            )
            ax.add_patch(rect)

        plt.show()
